/*
 * (C) Copyright 2007, 2011 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#ifndef I2C00_H
#define I2C00_H

#include <config.h>
#include <asm/io.h>

/********************************************************/
/* User configurations                                  */
/********************************************************/
/* PSCLK */
#define PSCLK_NSEC              (1000/CONFIG_I2C00_CLOCK)      /* nsec (=1000*1/MHz) */

/* I2C00 BASE ADDRESS */
#define I2C00_BASE_00           (0x51012000)
#define I2C00_BASE_01           (0x30003000)
#define I2C00_BUS_MAX           (2)

#define I2C00_SLAVEADDR_DEFAULT (0)

/* Platform infomation */
#define I2C00_RAM_OFFSET        0x1000  /* Offset from DEFAULT_BASE */

/* wait ack timeout */
#define I2C00_WAIT_ACK_TIMEOUT	(2)	/* 2sec */

/********************************************************/
/* Offset address of Registers                          */
/********************************************************/
#define ICDA            (0x000)         /* data shift register */
#define ICAR            (0x000)         /* address register */
#define ICCS            (0x004)         /* control status register */
#define ICCM            (0x008)         /* mode register */
#define ICCCL           (0x00c)         /* clock control register low */
#define ICCCH           (0x010)         /* clock control register high */
#define ICATMC          (0x014)         /* auto transfer control register */
#define ICTCR           (0x018)         /* auto transfer byte register */
#define ICATMS          (0x01c)         /* auto transfer status register */
#define ICDS            (0x020)         /* data setup control register */
#define ICINT           (0x024)         /* interrupt status register */
#define ICINTC          (0x028)         /* interrupt clear register */
#define ICBE            (0x02c)         /* dummy-i/o register */
#define ICAL            (0x030)         /* arbitration status register */
#define ICID            (0x0fc)         /* ID register */

#define ICRAM           (I2C00_RAM_OFFSET) /* RAM top offset address */
#define ICRAMSIZE       (64)            /* RAM buffer byte size */

/********************************************************/
/* Bit assign                                           */
/********************************************************/
/* ICCS */
#define MST             (0x80)
#define TRX             (0x40)
#define BB              (0x20)
#define PIN             (0x10)
#define ES0             (0x08)
#define AAS             (0x04)
#define AD0             (0x02)
#define LRB             (0x01)
#define BC2             (0x04)
#define BC1             (0x02)
#define BC0             (0x01)

/* ICCM */
#define ACK             (0x80)
#define ALMSK           (0x40)
#define EXADR           (0x08)
#define ATM             (0x04)
#define SDA             (0x02)
#define SCL             (0x01)

/* ICATMC */
#define IRST            (0x80)
#define ERRST           (0x40)
#define EXT             (0x10)
#define APC             (0x04)
#define AT1             (0x02)
#define AT0             (0x01)

/* ICATMS */
#define AS2             (0x04)
#define AS1             (0x02)
#define AS0             (0x01)

/* ICINT */
#define I2CSEN          (0x40)
#define ATMEN           (0x20)
#define I2CEN           (0x10)
#define INTI2CS         (0x04)
#define INTATM          (0x02)
#define INTI2C          (0x01)

/* ICINTC */
#define INTI2CSC        (0x04)
#define INTATMC         (0x02)
#define INTI2CC         (0x01)

/* ICBE */
#define TSDAIE          (0x80)
#define TSCLIE          (0x40)
#define TSDAOE          (0x20)
#define TSCLOE          (0x10)
#define TSDAI           (0x08)
#define TSCLI           (0x04)
#define TSDAO           (0x02)
#define TSCLO           (0x01)

/* ICAL */
#define AL              (0x01)

/********************************************************/
/* Macro                                                */
/********************************************************/
/* Transfer rate (for u-boot I/F) */
#define I2C00_SPEED_100KBPS		(100000)
#define I2C00_SPEED_400KBPS		(400000)

/* slave address 0bit(R/W) */
#define I2C00_RW_READ			(0x01)
#define I2C00_RW_WRITE			(~I2C00_RW_READ)

/* FIRST->idle , NORMAL->hold */
#define I2C00_HOLD			(0)
#define I2C00_FIRST			(1)

/* wait_ack() return value */
#define I2C00_E_OK			(1)
#define I2C00_E_NO_INT			(0)
#define I2C00_E_TIMEOUT			(-1)
#define I2C00_E_NOACK			(-2)
#define I2C00_E_ERR			(-3)
#define I2C00_E_FREEZE			(-4)

#define I2C00_BUF_LEN			(255)

/* register access */
#define RIN(n)		readl((g_base_addr[g_bus]+(n)))
#define ROUT(n, v)	writel((v), (g_base_addr[g_bus]+(n)))

/* Transfer rate */
#define I2C00_100KBPS           (0)             /* Standard mode (100kbps) */
#define I2C00_400KBPS           (1)             /* Fast mode (400kbps) */

#define I2C00_SEMI              (AT0)           /* Semi auto (ST/Sr + data) */
#define I2C00_SP                (AT1)           /* Stop condition (SP) */
#define I2C00_FULL              (AT1 | AT0)     /* Full auto (ST/Sr + data + SP) */

/* Timing parameter,  STANDARD-MODE(100kHz) */
#define STD_THIGH_MUL1000       (4300000 - STD_TF_MUL1000)      /* SCL clock high, hold time(ns) * 1000 */
#define STD_TLOW_MUL1000        (5700000 - STD_TR_MUL1000)      /* SCL clock low, hold time(ns) * 1000 */
#define STD_TR_MUL1000          (0)				/* SCL up-edge time(ns) * 1000 */
#define STD_TF_MUL1000          (0)				/* SCL down-edge time(ns) * 1000 */
#define STD_VIHMIN_MUL1000      (700)           /* VIHmin(vdd) * 1000 */
#define STD_VILMAX_MUL1000      (300)           /* VILmin(vdd) * 1000 */
#define STD_TSUDAT_MUL1000      (250000)        /* data setup time(ns) * 1000 */
#define STD_LCNT_MUL1000        (STD_TLOW_MUL1000 - \
                                ((STD_TR_MUL1000 * (STD_VILMAX_MUL1000 - 0)) / \
                                (STD_VIHMIN_MUL1000 - STD_VILMAX_MUL1000)))     /* Low count time * 1000 */
#define STD_HCNT_MUL1000        (STD_THIGH_MUL1000 - \
                                ((STD_TF_MUL1000 * (1000 - STD_VIHMIN_MUL1000)) / \
                                (STD_VIHMIN_MUL1000 - STD_VILMAX_MUL1000)))     /* High count time * 1000 */

#define STD_LCNT                (STD_LCNT_MUL1000 / 1000)
#define STD_HCNT                (STD_HCNT_MUL1000 / 1000)
#define STD_TR_TSUDAT           ((STD_TR_MUL1000 + STD_TSUDAT_MUL1000) / 1000)

/* Timing parameter,  FAST-MODE(400kHz) */
#define FAST_THIGH_MUL1000      (900000  - FAST_TF_MUL1000)        /* SCL clock high, hold time(ns) * 1000 */
#define FAST_TLOW_MUL1000       (1600000 - FAST_TR_MUL1000)        /* SCL clock low, hold time(ns) * 1000 */
#define FAST_TR_MUL1000         (0)        /* SCL up-edge time(ns) * 1000 */
#define FAST_TF_MUL1000         (0)        /* SCL down-edge time(ns) * 1000 */
#define FAST_VIHMIN_MUL1000     (700)           /* VIHmin(vdd) * 1000 */
#define FAST_VILMAX_MUL1000     (300)           /* VILmin(vdd) * 1000 */
#define FAST_TSUDAT_MUL1000     (100000)        /* data setup time(ns) * 1000 */
#define FAST_LCNT_MUL1000       (FAST_TLOW_MUL1000 - \
                                ((FAST_TR_MUL1000 * (FAST_VILMAX_MUL1000 - 0)) / \
                                (FAST_VIHMIN_MUL1000 - FAST_VILMAX_MUL1000)))   /* Low count time * 1000 */
#define FAST_HCNT_MUL1000       (FAST_THIGH_MUL1000 - \
                                ((FAST_TF_MUL1000 * (1000 - FAST_VIHMIN_MUL1000)) / \
                                (FAST_VIHMIN_MUL1000 - FAST_VILMAX_MUL1000)))   /* High count time * 1000 */
#define FAST_LCNT               (FAST_LCNT_MUL1000 / 1000)
#define FAST_HCNT               (FAST_HCNT_MUL1000 / 1000)
#define FAST_TR_TSUDAT          ((FAST_TR_MUL1000 + FAST_TSUDAT_MUL1000) / 1000)

/* Configuration Macro */
#define I2C00_CFG_ICDH          (((PSCLK_NSEC *  7) >= 300) ? 0 : \
                                 ((PSCLK_NSEC * 15) >= 300) ? 1 : \
                                 ((PSCLK_NSEC * 23) >= 300) ? 2 : 3)

#define ADDJUST                 1  /* 0:down round 1:up round */

#define GET_ICCCL(rate)         ({ \
                                  unsigned long _lval = \
                                    (((rate) == I2C00_400KBPS) ? \
                                    (unsigned long)(FAST_LCNT / PSCLK_NSEC) : \
                                    (unsigned long)(STD_LCNT  / PSCLK_NSEC)) + ADDJUST; \
                                  ((_lval < 6) ? 6 : _lval); \
                                })

#define GET_ICCCH(rate)         ({ \
                                  unsigned long _lval = \
                                    (((rate) == I2C00_400KBPS) ? \
                                    (unsigned long)(FAST_HCNT / PSCLK_NSEC) : \
                                    (unsigned long)(STD_HCNT  / PSCLK_NSEC)) + ADDJUST; \
                                  ((_lval < 6) ? 6 : _lval); \
                                })

#define GET_ICDS(rate)          ({ \
                                  unsigned long _lval = \
                                    (((rate) == I2C00_400KBPS) ? \
                                    (unsigned long)(FAST_TR_TSUDAT / PSCLK_NSEC) : \
                                    (unsigned long)(STD_TR_TSUDAT  / PSCLK_NSEC)) + ADDJUST; \
                                  ((_lval == 0) ? 1 : _lval); \
                                })

#endif  /* I2C00_H */
