/*
 * (C) Copyright 2007, 2011 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */


#include <common.h>

#if defined(CONFIG_XARINA)

#include "i2c00.h"


#define mmio_write32(address, value)         \
    *((volatile uint32_t *)address) = value;

#define mmio_read32(addr) *((volatile uint32_t *)(addr))

#if 0
#define I2C_ICCCL__VAL_100KBPS      32
#define I2C_ICCCH__VAL_100KBPS      32
#define I2C_ICCM_ICDH__VAL_100KBPS  1
#define I2C_ICDS__VAL_100KBPS       9
#else
#define I2C_ICCCL__VAL_100KBPS      GET_ICCCL(I2C00_100KBPS)
#define I2C_ICCCH__VAL_100KBPS      GET_ICCCH(I2C00_100KBPS)
#define I2C_ICCM_ICDH__VAL_100KBPS  I2C00_CFG_ICDH
#define I2C_ICDS__VAL_100KBPS       GET_ICDS(I2C00_100KBPS)
#endif

#define I2C_SLV_ADDR_EEPROM         (0x50)	/* SerialEEPROM */

#if defined(CONFIG_XARINA_VSS)
  #define I2C0                I2C00_BASE_01		/* I2C-1 */
#else
  #define I2C0                I2C00_BASE_00		/* I2C-0 */
#endif

#define I2C_RAM_TOP(x)      (x + I2C00_RAM_OFFSET)
#define I2C_RAM_BOT(x)      (x + I2C00_RAM_OFFSET + 0x100)

#define I2C_ICDA(x)         (x + ICDA)
#define I2C_ICAR(x)         (x + ICAR)
#define I2C_ICCS(x)         (x + ICCS)
#define I2C_ICCM(x)         (x + ICCM)
#define I2C_ICCCL(x)        (x + ICCCL)
#define I2C_ICCCH(x)        (x + ICCCH)
#define I2C_ICATMC(x)       (x + ICATMC)
#define I2C_ICTCR(x)        (x + ICTCR)
#define I2C_ICATMS(x)       (x + ICATMS)
#define I2C_ICDS(x)         (x + ICDS)
#define I2C_ICINT(x)        (x + ICINT)
#define I2C_ICINTC(x)       (x + ICINTC)
#define I2C_ICBE(x)         (x + ICBE)
#define I2C_ICAL(x)         (x + ICAL)
#define I2C_ICID(x)         (x + ICID)

/* I2C EEPROM DDR parameter struct */
typedef struct {
    struct {
        uint32_t base_addr;
        uint32_t num_of_data;
    } hdr;
    uint32_t data[0];
} I2C_EEPROM_DDR_PARAM;

#define I2C_EEPROM_DDR_PARAM_BUF_SIZE   56
typedef struct {
    union {
        I2C_EEPROM_DDR_PARAM s;
        uint32_t array[I2C_EEPROM_DDR_PARAM_BUF_SIZE];
    };
} DDR_PARAM_BUF;


static void se2p_i2c_reset(void)
{
    /* set ICCCL to zero */
    mmio_write32(I2C_ICCCL(I2C0), 0);

    /* ICATMC (issue software reset) */
    mmio_write32(I2C_ICATMC(I2C0), 0x80);

    /* dummy read */
    mmio_read32(I2C_ICCCL(I2C0));
}

void se2p_i2c_init(void)
{
    se2p_i2c_reset();

    /* mask out all interrupts */
    mmio_write32(I2C_ICINT(I2C0), 0);

    /* normal transfer mode, 100Kbps, 7bit-slv-addr, 
     * disable ack-bit, enable arbitration lost detection */
    mmio_write32(I2C_ICCCL(I2C0), I2C_ICCCL__VAL_100KBPS);
    mmio_write32(I2C_ICCCH(I2C0), I2C_ICCCH__VAL_100KBPS);
    mmio_write32(I2C_ICDS(I2C0), I2C_ICDS__VAL_100KBPS);
    mmio_write32(I2C_ICCM(I2C0), (I2C_ICCM_ICDH__VAL_100KBPS << 8));

    /* to enable I2C bus control, set ICCS ESO to 1 */
    mmio_write32(I2C_ICCS(I2C0), 0x08);
}

static void se2p_i2c_eeprom_set_addr(uint32_t src)
{
    /* ICTCR (transfer byte counter) */
    mmio_write32(I2C_ICTCR(I2C0), 2);

    /* set I2C slave address */
    uint32_t *i2c_ram_ptr = (uint32_t *)I2C_RAM_TOP(I2C0);
    uint8_t slv_addr;
    slv_addr = I2C_SLV_ADDR_EEPROM << 1;                 /* I2C slave address */
    mmio_write32(i2c_ram_ptr, slv_addr);
    i2c_ram_ptr++;

    /* set transfer data */
    mmio_write32(i2c_ram_ptr, (src >> 8) & 0x0fUL);
    i2c_ram_ptr++;
    mmio_write32(i2c_ram_ptr, src & 0x0ffUL);
    i2c_ram_ptr++;

    /* To set controller auto-transfer mode, set ICCM ATM to 1 */
    uint32_t i2c_iccm_val = mmio_read32(I2C_ICCM(I2C0));
    i2c_iccm_val |= 0x04;
    mmio_write32(I2C_ICCM(I2C0), i2c_iccm_val);

    /* To kick controller in fullauto-mode, set ICATMC APC to 1, and AT0/1 to 11b */
    mmio_write32(I2C_ICATMC(I2C0), 0x07);

    /* wait for completion of transfer */
    uint32_t i2c_icintc_val = mmio_read32(I2C_ICINT(I2C0));
    while (0x0 == (i2c_icintc_val & 0x07)) i2c_icintc_val = mmio_read32(I2C_ICINT(I2C0));
    mmio_write32(I2C_ICINTC(I2C0), i2c_icintc_val);
}

void se2p_i2c_eeprom_rd(uint32_t dst, uint32_t src, uint32_t size)
{
    uint32_t rem_size = size;

    while (0 != rem_size) {
        uint32_t pg_bound_limit_size = (0x20UL - (src % 0x20));
        if (rem_size > pg_bound_limit_size) {
            size = pg_bound_limit_size;
        }
        else {
            size = rem_size;
        }

        /* is the state of IIC bus idle ? */
        {
            uint32_t i2c_iccs_val;
            i2c_iccs_val = mmio_read32(I2C_ICCS(I2C0));
            if (0 != (i2c_iccs_val & 0x80)) {
		*(uint32_t *)dst = 0;
		return;
            }
        }

        se2p_i2c_eeprom_set_addr(src);

        /* ICTCR (transfer byte counter) */
        mmio_write32(I2C_ICTCR(I2C0), size);

        /* set I2C slave address */
        uint32_t *i2c_ram_ptr = (uint32_t *)I2C_RAM_TOP(I2C0);
        uint8_t slv_addr;
        slv_addr = I2C_SLV_ADDR_EEPROM << 1;             /* I2C slave address */
        slv_addr |= 0x01;                               /* read */
        mmio_write32(i2c_ram_ptr, slv_addr);
        i2c_ram_ptr++;

        /* To set controller auto-transfer mode, set ICCM ATM to 1 */
        uint32_t i2c_iccm_val = mmio_read32(I2C_ICCM(I2C0));
        i2c_iccm_val |= 0x04;
        mmio_write32(I2C_ICCM(I2C0), i2c_iccm_val);

        /* To kick controller in fullauto-mode, set ICATMC APC to 1, and AT0/1 to 11b */
        mmio_write32(I2C_ICATMC(I2C0), 0x07);

        /* wait for completion of transfer */
        uint32_t i2c_icintc_val = mmio_read32(I2C_ICINT(I2C0));
        while (0x0 == (i2c_icintc_val & 0x07)) i2c_icintc_val = mmio_read32(I2C_ICINT(I2C0));
        uint32_t i2c_icatms_val = mmio_read32(I2C_ICATMS(I2C0));
        mmio_write32(I2C_ICINTC(I2C0), i2c_icintc_val);

        /* check if transfer was succeeded or not */
        if (0x0 != (i2c_icatms_val & 0x07)) {
		*(uint32_t *)dst = 0;
		return;
        }

        /* copy data from I2C RAM buffer */
        {
            uint32_t loop;
            uint8_t *src_ptr = (uint8_t *)i2c_ram_ptr;
            uint8_t *dst_ptr = (uint8_t *)dst;
            for (loop = 0; loop < size; loop++) {
                *dst_ptr = *src_ptr;
                dst_ptr++;
                src_ptr += 4;
            }
        }

        src += size;
        dst += size;
        rem_size -= size;
    }

    return;
}

#endif /* defined(CONFIG_XARINA)*/

