/*
 * (C) Copyright 2007, 2011 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>

#if defined(CONFIG_CMD_NAND) && !defined(CONFIG_NAND_LEGACY)

#include <command.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/mtd.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/arch/xarina_nand2.h>
#include <asm/arch/xarina_dmac.h>
#ifdef CONFIG_POSTEDWRITE_WORKAROUND
#include <asm/arch/postedwrite_sync.h>
#endif
#if !defined(CONFIG_NAND_SPL)
#define CONFIG_MTD_XARINA_NAND_DMAC
#endif

#define XARINA_NANDIF_PHYS_BASE		(0x38006000)

/* 16bit access macro for xarina */
#define WRITE16	writew
#define READ16	readw

#define XARINA_READID_ADDR	0
#define XARINA_READ_BYTE_CNT_MAX	6
#define XARINA_WORD_MASK		0x00000003


/**************************/
/* prototype              */
/**************************/
static int xarina_nand2_check_task(struct mtd_info *mtd, unsigned short task);
static void xarina_nand2_set_task(struct mtd_info *mtd, unsigned short task);
void xarina_nand2_write_protect(struct mtd_info *mtd, int protect);

static struct s_xarina_nand_info {

	unsigned char	cmp_buf[XARINA_NAND2_PAGESIZE];
	unsigned char	status_id_read_cnt;
	unsigned char	read_byte_reg[6];
	unsigned short	ecc_task;
	unsigned short	check_ecc_buf;

} xa_nand_info;

#if defined(CONFIG_MTD_XARINA_NAND_DMAC)
 __attribute__ ((aligned (32)))
static unsigned char dmac_buf[XARINA_NAND2_ECC_SIZE];
#endif

#if !defined(CONFIG_NAND_SPL)
/***********************/
/* internal variable   */
/***********************/
/* for autooob */
#if 0
static struct nand_oobinfo nand_oob_128_1bit = {
	.useecc = MTD_NANDECC_AUTOPLACE,
	.eccbytes = 8,
	.eccpos = { 120, 121, 122, 123, 124, 125, 126, 127 },
	.oobfree = { {1, 120}, },
};
#endif


/* XARINA_NAND2 ECCLAYOUT : ecc 10bytes for 512 bytes or 8bytes */
static struct nand_ecclayout xarina_nand2_oob = {/* for 2GB */
	.eccbytes = XARINA_NAND2_ECC_TOTAL + XARINA_NAND2_OOB_ECC_TOTAL,
	.eccpos = {  28, 29, 30, 31, 32, 33, 34, 35, 36, 37,
		     38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
		     48, 49, 50, 51, 52, 53},
	.oobfree = { {0, XARINA_NAND2_OOBSIZE - XARINA_NAND2_ECC_TOTAL - XARINA_NAND2_OOB_ECC_TOTAL}, },
};

/* for bbt */
static uint8_t bbt_pattern[] = {'B', 'b', 't', '0' };
static uint8_t mirror_pattern[] = {'1', 't', 'b', 'B' };

/* main discriptor */
static struct nand_bbt_descr xarina_bbt_main_descr = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE |
			NAND_BBT_WRITE | NAND_BBT_2BIT |
			NAND_BBT_VERSION | NAND_BBT_PERCHIP |
			NAND_BBT_SCAN2NDPAGE |NAND_NO_SUBPAGE_WRITE,
	.offs = 1,		// BBT pattern offset
	.len = 4,		// BBT pattern length
	.veroffs = 0,		// version location
	.maxblocks = 4,		// btt location max block
	.pattern = bbt_pattern	// bbt pattern
};

/* mirror discriptor */
static struct nand_bbt_descr xarina_bbt_mirror_descr = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE |
			NAND_BBT_WRITE | NAND_BBT_2BIT |
			NAND_BBT_VERSION | NAND_BBT_PERCHIP |
			NAND_BBT_SCAN2NDPAGE |NAND_NO_SUBPAGE_WRITE,
	.offs =	1,		// BBT pattern offset
	.len = 4,		// BBT pattern length
	.veroffs = 0,		// version location
	.maxblocks = 4,		// btt location max block
	.pattern = mirror_pattern	// bbt pattern
};

static uint8_t scan_ff_pattern[] = { 0xff };

static struct nand_bbt_descr xarina_largepage_flashbased = {
	.options = NAND_BBT_SCAN2NDPAGE | NAND_BBT_SCANEMPTY,
	.offs = 0,		// bad block mark offset
	.len = 1,		// bad block mark length
	.pattern = scan_ff_pattern	// good block mark pattern
};




/******************************************/
/* function for replace or board specific */
/******************************************/


/**
 * xarina_nand2_write_cmd
 *
 * write command function
 * cmd1,2,3 : -1 indicates invalid
*/
static void xarina_nand2_write_cmd(struct mtd_info *mtd, 
	unsigned int cmd1, unsigned int cmd2, unsigned int cmd3)
{
	unsigned short tmp_cmd;
	register struct nand_chip *chip = mtd->priv;

	/* cmd2 & cmd3 are optional */
	if (cmd1 != XARINA_NAND2_CMD_INVALID) {
		tmp_cmd = cmd1 & XARINA_NAND2_MASK_8BIT;
		/* set command1 */
		WRITE16(tmp_cmd, chip->IO_ADDR_W + XARINA_NAND2_COMMAND1_REG);

	}
	if (cmd2 != XARINA_NAND2_CMD_INVALID) {
		tmp_cmd = cmd2 & XARINA_NAND2_MASK_8BIT;
		/* set command2 */
		WRITE16(tmp_cmd, chip->IO_ADDR_W + XARINA_NAND2_COMMAND2_REG);
	}

	if (cmd3 != XARINA_NAND2_CMD_INVALID) {
		tmp_cmd = cmd3 & XARINA_NAND2_MASK_8BIT;
		/* set command3 */
		WRITE16(tmp_cmd, chip->IO_ADDR_W + XARINA_NAND2_COMMAND3_REG);
	}
	if (cmd1 == NAND_CMD_READID) {  
		tmp_cmd = READ16(chip->IO_ADDR_W + XARINA_NAND2_MODE_REG);
		tmp_cmd &= ~XARINA_NAND2_BIT_REG_NUM;
		tmp_cmd |= XARINA_NAND2_BIT_READ_ID;
		WRITE16(tmp_cmd, chip->IO_ADDR_W + XARINA_NAND2_MODE_REG);
		xa_nand_info.status_id_read_cnt = 0;
	}
	if (cmd1 == NAND_CMD_STATUS) {
		tmp_cmd = READ16(chip->IO_ADDR_W + XARINA_NAND2_MODE_REG);
		tmp_cmd &= ~XARINA_NAND2_BIT_REG_NUM;
		tmp_cmd |= XARINA_NAND2_BIT_READ_STATUS;
		WRITE16(tmp_cmd, chip->IO_ADDR_W + XARINA_NAND2_MODE_REG);
		xa_nand_info.status_id_read_cnt = 0;
	}

	return;
}

/**
 * xarina_nand2_write_addr
 *
 * write address function
 * column, page_addr : -1 indicates invalid
*/
static void xarina_nand2_write_addr(struct mtd_info *mtd, int column, int page_addr)
{

	unsigned long tmp_addr;
	register struct nand_chip *chip = mtd->priv;

	if(page_addr != XARINA_NAND2_ADDR_INVALID){ /* page_addr must be valid */

		tmp_addr = (u_short)(page_addr & XARINA_NAND2_MASK_16BIT);

		/* set page0, page1 */
		WRITE16(tmp_addr, chip->IO_ADDR_W + XARINA_NAND2_PAGE_ADDR01_REG);

		tmp_addr = (u_short)((page_addr >> XARINA_NAND2_SHIFT_16) & XARINA_NAND2_MASK_8BIT);
		/* set page2 */
		WRITE16(tmp_addr, chip->IO_ADDR_W + XARINA_NAND2_PAGE_ADDR2_REG);

	}

	if (column != XARINA_NAND2_ADDR_INVALID) {

		tmp_addr = (u_short)(column & XARINA_NAND2_MASK_16BIT);

		/* set col0, col1 */
		WRITE16(tmp_addr, chip->IO_ADDR_W + XARINA_NAND2_COL_ADDR_REG);

	}

	return;
}

/**
 * xarina_nand2_command
 * @mtd:        MTD device structure
 * @command:    nand_command
 * @column:     offset inside page
 * @page_addr:  page offset
 *
 */
static void xarina_nand2_command(struct mtd_info *mtd, 
			unsigned int command, int column, int page_addr) 
{

	register struct nand_chip *chip = mtd->priv;
	unsigned short task;


	switch (command) {
	/* Emulate NAND_CMD_READOOB */
	case NAND_CMD_READOOB:
	case NAND_CMD_READ0:

		/* command setting */
		xarina_nand2_write_cmd(mtd, command, NAND_CMD_READSTART,
					XARINA_NAND2_CMD_INVALID);
		/* address setting */
		xarina_nand2_write_addr(mtd, column, page_addr);

		task = XARINA_NAND2_BIT_TASK_COMMAND1
				| XARINA_NAND2_BIT_TASK_COMMAND2
				| XARINA_NAND2_BIT_TASK_COL_ADR0
				| XARINA_NAND2_BIT_TASK_COL_ADR1
				| XARINA_NAND2_BIT_TASK_PAGE_ADR0
				| XARINA_NAND2_BIT_TASK_PAGE_ADR1
				| XARINA_NAND2_BIT_TASK_PAGE_ADR2
				| XARINA_NAND2_BIT_TASK_WAIT_READY1;

		xarina_nand2_set_task(mtd,task);

		/* ready waiting */
		while (1) {
			if (chip->dev_ready(mtd)) {
				break;
			}
		}

		break;

	case NAND_CMD_READID:
		/* command setting */
		xarina_nand2_write_cmd(mtd, command, XARINA_NAND2_CMD_INVALID, XARINA_NAND2_CMD_INVALID);
		/* address setting */
		xarina_nand2_write_addr(mtd, XARINA_NAND2_ADDR_INVALID, XARINA_READID_ADDR);
		task = XARINA_NAND2_BIT_TASK_COMMAND1
			| XARINA_NAND2_BIT_TASK_PAGE_ADR0
			| XARINA_NAND2_BIT_TASK_RECV2REG;
		xarina_nand2_set_task(mtd,task);
		break;

	case NAND_CMD_STATUS:
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_OFF);
		/* command setting */
		xarina_nand2_write_cmd(mtd, command, 
			XARINA_NAND2_CMD_INVALID, XARINA_NAND2_CMD_INVALID);

		task = XARINA_NAND2_BIT_TASK_COMMAND1
				| XARINA_NAND2_BIT_TASK_RECV2REG;

		xarina_nand2_set_task(mtd,task);
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_ON);
		break;

	case NAND_CMD_SEQIN:
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_OFF);
		/* command setting */
		xarina_nand2_write_cmd(mtd, command, 
				XARINA_NAND2_CMD_INVALID, XARINA_NAND2_CMD_INVALID);

		/* address setting */
		xarina_nand2_write_addr(mtd, column, page_addr);

		task = XARINA_NAND2_BIT_TASK_COMMAND1
				| XARINA_NAND2_BIT_TASK_COL_ADR0
				| XARINA_NAND2_BIT_TASK_COL_ADR1
				| XARINA_NAND2_BIT_TASK_PAGE_ADR0
				| XARINA_NAND2_BIT_TASK_PAGE_ADR1
				| XARINA_NAND2_BIT_TASK_PAGE_ADR2
				| XARINA_NAND2_BIT_TASK_WAIT_READY1;

		xarina_nand2_set_task(mtd,task);

		break;

	case NAND_CMD_PAGEPROG:

		/* command setting */
		xarina_nand2_write_cmd(mtd, command,
				XARINA_NAND2_CMD_INVALID, XARINA_NAND2_CMD_INVALID);

		task = XARINA_NAND2_BIT_TASK_COMMAND1
				| XARINA_NAND2_BIT_TASK_WAIT_READY1;
		xarina_nand2_set_task(mtd,task);

		/* ready waiting */
		while (1) {
			if (chip->dev_ready(mtd)) {
				break;
			}
		}
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_ON);

		break;
	case NAND_CMD_ERASE1:

		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_OFF);

		/* command setting */
		xarina_nand2_write_cmd(mtd, command,
				XARINA_NAND2_CMD_INVALID, XARINA_NAND2_CMD_INVALID);

		/* address setting */
		xarina_nand2_write_addr(mtd, XARINA_NAND2_ADDR_INVALID, page_addr);

		task = XARINA_NAND2_BIT_TASK_COMMAND1
				| XARINA_NAND2_BIT_TASK_PAGE_ADR0
				| XARINA_NAND2_BIT_TASK_PAGE_ADR1
				| XARINA_NAND2_BIT_TASK_PAGE_ADR2
				| XARINA_NAND2_BIT_TASK_WAIT_READY1;

		xarina_nand2_set_task(mtd,task);

		break;

	case NAND_CMD_ERASE2:

		/* wait for all task finished */
		while (1) {
			if (xarina_nand2_check_task(mtd, XARINA_NAND2_BIT_TASK_ALL) == 0)
				break;
		}

		/* command setting */
		xarina_nand2_write_cmd(mtd, command,
				XARINA_NAND2_CMD_INVALID, XARINA_NAND2_CMD_INVALID);

		task = XARINA_NAND2_BIT_TASK_COMMAND1
				| XARINA_NAND2_BIT_TASK_WAIT_READY1;

		xarina_nand2_set_task(mtd,task);

		/* ready waiting */
		while (1) {
			if (chip->dev_ready(mtd)) {
        		        break;
			}
		}
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_ON);

		break;
		/* NAND_CMD_ERASE2 -> NAND_CMD_RESET sequence */

	case NAND_CMD_RESET:
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_OFF);
		/* command setting */
		xarina_nand2_write_cmd(mtd, NAND_CMD_RESET,
			XARINA_NAND2_CMD_INVALID, XARINA_NAND2_CMD_INVALID);

		task = XARINA_NAND2_BIT_TASK_COMMAND1
				| XARINA_NAND2_BIT_TASK_WAIT_READY1;
		xarina_nand2_set_task(mtd,task);


		/* ready waiting */
		while (1) {
			if (chip->dev_ready(mtd)) {
				break;
			}
		}
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_ON);
		break;

	default:
		break;
	}

#if !defined(CONFIG_NAND_SPL)
	ndelay(100);
	nand_wait_ready(mtd);
#endif

	return;
}


/*
 *	hardware specific access to control-lines
 *	NOTE:  boards may use different bits for these!!
 *
 *	ctrl:
 *	NAND_NCE: bit 0 -  ~ChipEnable
 *	NAND_CLE: bit 1 -> bit 1  (0x0002)
 *	NAND_ALE: bit 2 -> bit 2  (0x0004)
 *
 *	NAND_CTRL_CHANGE: bit 7 -> bit 7  (0x0080)
 */

static void xarina_cmd_ctrl(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	/* dummy funciton */

	return;
}



/*****************************************************************************/
/*****************************************************************************/
/*  READ OPERATION                                                           */
/*****************************************************************************/
/*****************************************************************************/
static uint8_t xarina_read_byte(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	uint8_t ret;
	uint16_t tmp_buf;

	if(xa_nand_info.status_id_read_cnt >= XARINA_READ_BYTE_CNT_MAX){
		return 0;
	}

	if (xa_nand_info.status_id_read_cnt == 0) {
		tmp_buf = READ16(chip->IO_ADDR_R + XARINA_NAND2_STATUS_ID_READ12_REG);
		xa_nand_info.read_byte_reg[0] = (uint8_t)(tmp_buf & XARINA_NAND2_MASK_8BIT);
		xa_nand_info.read_byte_reg[1] = (uint8_t)((tmp_buf >> 8) & XARINA_NAND2_MASK_8BIT);
		tmp_buf = READ16(chip->IO_ADDR_R + XARINA_NAND2_STATUS_ID_READ34_REG);
		xa_nand_info.read_byte_reg[2] = (uint8_t)(tmp_buf & XARINA_NAND2_MASK_8BIT);
		xa_nand_info.read_byte_reg[3] = (uint8_t)((tmp_buf >> 8) & XARINA_NAND2_MASK_8BIT);
		tmp_buf = READ16(chip->IO_ADDR_R + XARINA_NAND2_STATUS_ID_READ56_REG);
		xa_nand_info.read_byte_reg[4] = (uint8_t)(tmp_buf & XARINA_NAND2_MASK_8BIT);
		xa_nand_info.read_byte_reg[5] = (uint8_t)((tmp_buf >> 8) & XARINA_NAND2_MASK_8BIT);

	}

	ret = xa_nand_info.read_byte_reg[xa_nand_info.status_id_read_cnt];

	xa_nand_info.status_id_read_cnt++;

	return ret;
}

static u16 xarina_read_word(struct mtd_info *mtd)
{
#if !defined(CONFIG_NAND_SPL)
	printk("nand_read_word not supported\n");
#endif
	return 0;
}



/**
 * xarina_nand2_read_page_raw
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	buffer to store read data
 */
static int xarina_nand2_read_page_raw(struct mtd_info *mtd, 
			struct nand_chip *chip, uint8_t *buf, int page)
{

	int i, j;
	uint8_t *data_offset;
	uint8_t *oob_offset;
	uint8_t *ecc_offset;

/*  | <---------------- read data 1 step -------------------> |             */
/*  ,-------------------------------------------------+-------+             */
/*  |            data(512)                            |ecc(10)|             */
/*  `-------------------------------------------------+-------+             */

/*  | read oob 1 step   |             */
/*  ,---------+---------+             */
/*  | oob(8)  | ecc(10) |             */
/*  `---------+---------+             */

/*        
*  | | 1. read data 512 byte to ( buf + i * 512 + 2048 * j)
*  | | 2. read ecc 10 byte to (oob_pointer + ECC_OFFSET + i * 10 + j * 40)
*  | ---- repeat 4 times -----
*  | 3. read oob 8 byte to (oob_pointer + j * 8 )
*  | 4. read ecc 10 byte to (oob_pointer + ECC_OFFSET + ECC_TOTAL - (2-j) * 10)
*  ----- repeat 2 times -----
*   5. read reserved area 12byte
*
*/
	for(j = 0; j < 2; j++) {
		/* ecc.steps=8 , 4 times loop*/
		for(i = 0; i < chip->ecc.steps / 2; i++) {

			/* data_offset (512+10)*i + 2048 * j */
			/* read data = ecc.size = 512 byte */
			data_offset = buf + (i * chip->ecc.size) + XARINA_NAND2_PAGESIZE / 2 * j ;
			chip->read_buf(mtd, data_offset, chip->ecc.size);

			/* oobsize=128, ecc.steps=8 ecc.bytes=10 */
			ecc_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_ECC_START_OFFSET + chip->ecc.bytes * (i + 4 * j));
			/* read ecc (10 byte) */
			chip->read_buf(mtd, ecc_offset, chip->ecc.bytes);

		}

		/* OOB:8bytes */
		oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * j);
		chip->read_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);

		ecc_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOBSIZE - (2 - j) * chip->ecc.bytes);
		/* read ecc (10 byte) */
		chip->read_buf(mtd, ecc_offset, chip->ecc.bytes);
 
	}
	
	/* reserved area 12bytes */
	oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * 2 );
	/* read reserved area (12 byte) */
	chip->read_buf(mtd, oob_offset, XARINA_NAND2_RESERVED_BYTES);


	return 0;
}

/**
 * xarina_nand2_read_page_hwecc
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	buffer to store read data
 *
 * Not for syndrome calculating ecc controllers which need a special oob layout
 */
static int xarina_nand2_read_page_hwecc(struct mtd_info *mtd, 
			struct nand_chip *chip, uint8_t *buf, int page)
{
	int i, j;
	uint8_t *oob_offset;
	uint8_t *data_offset;
	int stat;


/*  | <---------------- read data 1 step -------------------> |             */
/*  ,-------------------------------------------------+-------+             */
/*  |            data(512)                            |ecc(10)|             */
/*  `-------------------------------------------------+-------+             */
/*                                                    |hw_read|             */
/*  | read oob 1 step   |             */
/*  ,---------+---------+             */
/*  | oob(8)  | ecc(10) |             */
/*  `---------+---------+             */
/*            | hw_read |             */


	/* ecc_switch on */
	chip->ecc.hwctl(mtd, NAND_ECC_READ);

	for(j = 0; j < 2; j++) {
		/* ecc.steps=8, 4times loop */
		for(i = 0; i < chip->ecc.steps / 2; i++) {

			/* ecc.size = 512 byte */
			/* read data (512 byte) */
			data_offset = buf + (i * chip->ecc.size) + XARINA_NAND2_PAGESIZE / 2 * j ;
			chip->read_buf(mtd, data_offset, chip->ecc.size);

			if(READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_CHECK_REG) & XARINA_NAND2_BIT_ECCCHK){

				stat = chip->ecc.correct(mtd, data_offset, NULL, NULL);

				if (stat < 0)
					mtd->ecc_stats.failed++;
	 			else
	 				mtd->ecc_stats.corrected += stat;

			}

		}

		/* OOB:8bytes */
		oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * j);
		chip->read_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);

		if(READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_CHECK_REG) & XARINA_NAND2_BIT_ECCCHK){

			stat = chip->ecc.correct(mtd, oob_offset, NULL, NULL);
	
			if (stat < 0)
				mtd->ecc_stats.failed++;
			else
				mtd->ecc_stats.corrected += stat;
		}

	}

	xa_nand_info.ecc_task = 0;/* ecc_switch off */

	/* reserved area 12bytes */
	oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * 2 );
	/* read reserved area (12 byte) */
	chip->read_buf(mtd, oob_offset, XARINA_NAND2_RESERVED_BYTES);

	return 0;

}

/**
 * xarina_nand_read_oob_std
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @page:	page number to read
 * @sndcmd:	flag whether to issue read command or not
 */
static int xarina_nand2_read_oob_std(struct mtd_info *mtd, 
		struct nand_chip *chip, int page, int sndcmd)
{
	int col_offset;
	uint8_t *oob_offset;
	int stat;

	/* set offset, command */
	col_offset = (chip->ecc.size + chip->ecc.bytes) * chip->ecc.steps / 2;
	chip->cmdfunc(mtd, NAND_CMD_READ0, col_offset, page);

	chip->ecc.hwctl(mtd, NAND_ECC_READ);/* ecc_switch on */

	/* OOB:8bytes */
	oob_offset = (uint8_t *)(chip->oob_poi);
	chip->read_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);

	if(READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_CHECK_REG) & XARINA_NAND2_BIT_ECCCHK) {

		stat = chip->ecc.correct(mtd, oob_offset, NULL, NULL);

		if (stat < 0)
			mtd->ecc_stats.failed++;
		else
			mtd->ecc_stats.corrected += stat;

	}

	/* set offset, command */
	col_offset = (chip->ecc.size + chip->ecc.bytes) * chip->ecc.steps + XARINA_NAND2_OOB_UNIT + chip->ecc.bytes;
	chip->cmdfunc(mtd, NAND_CMD_READ0, col_offset, page);

	/* OOB:8bytes */
	oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT );
	chip->read_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);

	if(READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_CHECK_REG)&XARINA_NAND2_BIT_ECCCHK){

		stat = chip->ecc.correct(mtd, oob_offset, NULL, NULL);

		if (stat < 0)
			mtd->ecc_stats.failed++;
		else
			mtd->ecc_stats.corrected += stat;

	}

	xa_nand_info.ecc_task = 0;/* ecc_switch off */
			
	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
/*  WRITE OPERATION                                                          */
/*****************************************************************************/
/*****************************************************************************/

/**
 * xarina_nand2_write_page_raw
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	data buffer
 */
static void xarina_nand2_write_page_raw(struct mtd_info *mtd, 
		struct nand_chip *chip, const uint8_t *buf)
{
	int i, j;
	uint8_t *data_offset = NULL;
	uint8_t *ecc_offset = NULL;
	uint8_t *oob_offset = NULL;

	for(j = 0; j < 2; j++){
		for(i = 0; i < chip->ecc.steps / 2; i++){

			/* write data = ecc.size = 512 byte */
			data_offset = (uint8_t *)buf + (i*chip->ecc.size) + XARINA_NAND2_PAGESIZE / 2 * j ;
			chip->write_buf(mtd, data_offset, chip->ecc.size);

			/* oobsize=128, ecc.steps=8 ecc.bytes=10 */
			ecc_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_ECC_START_OFFSET + chip->ecc.bytes * (i + 4 * j));
			/* write oob to ecc area ( 10 byte ) */
			chip->write_buf(mtd, ecc_offset, chip->ecc.bytes);

		}
		/* OOB:8bytes */
		oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * j);
		chip->write_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);

		ecc_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOBSIZE - (2 - j) * chip->ecc.bytes);
		/* write ecc (10 byte) */
		chip->write_buf(mtd, ecc_offset, chip->ecc.bytes);

	}

	/* reserved area 12bytes */
	oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * 2);
	/* read reserved area (12 byte) */
	chip->write_buf(mtd, oob_offset, XARINA_NAND2_RESERVED_BYTES);

}

/**
 * xarina_nand2_write_page_hwecc
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	data buffer
 */
static void xarina_nand2_write_page_hwecc(struct mtd_info *mtd, 
			struct nand_chip *chip, const uint8_t *buf)
{
	int i, j;
	uint8_t *oob_offset;
	uint8_t *data_offset;

	chip->ecc.hwctl(mtd, NAND_ECC_WRITE);/* ecc_switch on */

	for(j = 0; j < 2; j++){

		for(i = 0; i < chip->ecc.steps / 2; i++){

			data_offset = (uint8_t *)buf + (i * chip->ecc.size) + XARINA_NAND2_PAGESIZE / 2 * j ;
			chip->write_buf(mtd, data_offset, chip->ecc.size);

		}

		/* OOB:8bytes */
		oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * j);
		chip->write_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);

	}

	xa_nand_info.ecc_task = 0;/* ecc_switch off */

	/* reserved area 12bytes */
	oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT * 2);
	/* read reserved area (12 byte) */
	chip->write_buf(mtd, oob_offset, XARINA_NAND2_RESERVED_BYTES);

}

/**
 * xarina_nand2_write_oob_std
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @page:	page number to write
 */
static int xarina_nand2_write_oob_std(struct mtd_info *mtd,
		struct nand_chip *chip, int page)
{
	int status = 0;
	int col_offset = 0;
	uint8_t *oob_offset;

	/* set offset, command */
	col_offset = (chip->ecc.size + chip->ecc.bytes) * chip->ecc.steps / 2;
	chip->cmdfunc(mtd, NAND_CMD_SEQIN, col_offset, page);


	/* OOB:8bytes */
	oob_offset = (uint8_t *)(chip->oob_poi);
	chip->ecc.hwctl(mtd, NAND_ECC_WRITE);/* ecc_switch on */
	chip->write_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);
	xa_nand_info.ecc_task = 0;/* ecc_switch off */
	chip->cmdfunc(mtd, NAND_CMD_PAGEPROG, -1, -1);

	status = chip->waitfunc(mtd, chip);
	if(status & NAND_STATUS_FAIL)
		return -EIO;

	/* set offset, command */
	col_offset = (chip->ecc.size + chip->ecc.bytes) * chip->ecc.steps + XARINA_NAND2_OOB_UNIT + chip->ecc.bytes;
	chip->cmdfunc(mtd, NAND_CMD_SEQIN, col_offset, page);

	/* OOB:8bytes */
	oob_offset = (uint8_t *)(chip->oob_poi + XARINA_NAND2_OOB_UNIT );
	chip->ecc.hwctl(mtd, NAND_ECC_WRITE);/* ecc_switch on */
	chip->write_buf(mtd, oob_offset, XARINA_NAND2_OOB_UNIT);
	xa_nand_info.ecc_task = 0;/* ecc_switch off */
	chip->cmdfunc(mtd, NAND_CMD_PAGEPROG, -1, -1);

	status = chip->waitfunc(mtd, chip);
	if(status & NAND_STATUS_FAIL)
		return -EIO;

	return 0;
}

/**
 * xarina_nand2_verify_buf - [DEFAULT] Verify chip data against buffer
 * @mtd:	MTD device structure
 * @buf:	buffer containing the data to compare
 * @len:	number of bytes to compare
 */
static int xarina_nand2_verify_buf(struct mtd_info *mtd,
					const uint8_t *buf, int len)
{
	int i;
	struct nand_chip *chip = mtd->priv;

	/* read data to cmp_buf */
	chip->ecc.read_page(mtd, chip, xa_nand_info.cmp_buf, 0);

	/* Compare Data */
	for (i = 0; i < len; i++){
		if(buf[i] != xa_nand_info.cmp_buf[i]){

			if(buf[i]!=0xff){ /* unless dummy_write */
				return -EFAULT;
			}
		}
	}
	return 0;
}


/*
 *  xarina_nand2_calculate_ecc
 *  Dummy Function
 */
static int xarina_nand2_calculate_ecc(struct mtd_info *mtd, 
				const u_char *dat, u_char *ecc_code)
{
	/* do nothing */
	return 0;
}

/**
 * xarina_nand2_correct_data
 * @mtd:        MTD device structure
 * @dat:        target data part
 * @oob:        target oob part
 * @ecc_data:   Stored ECC from NAND(don't use?)
 *
 * Correct Data of ECC
 */
static int xarina_nand2_correct_data(struct mtd_info *mtd,
			u_char *dat, u_char *oob, u_char *ecc_data)
{
	int error_corrected;
	struct nand_chip *chip = mtd->priv;
	u_short *e_addr;
	u_short e_bits;

	int ecc_num, i;
	u_short ecc_result;
	error_corrected = 0;

	/* calculation start */
	WRITE16(XARINA_NAND2_BIT_ECC_CALC_START, chip->IO_ADDR_W + XARINA_NAND2_ECC_CALC_START_REG);

	while(1){
		if(READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_CALC_START_REG) == 0)
			break;
	}

	/* Error Check */
	ecc_result = READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_RESULT_REG) >> XARINA_NAND2_ECC_RESULT_SHIFT;

	switch(ecc_result){
	case XARINA_NAND2_BIT_ECC_NO_ERR:
		/* No Error */
		return 0;
	case XARINA_NAND2_BIT_ECC_C_ERR:
		break;
	case XARINA_NAND2_BIT_ECC_UC_ERR:
	case XARINA_NAND2_BIT_ECC_5_ERR:
	default:
		return -1;
	}


	ecc_num = READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_RESULT_REG) & XARINA_NAND2_BIT_ECC_CORRECT_MASK;

	for(i = 0; i < ecc_num; i++){
		e_addr = (u_short *)(dat + 
				READ16(chip->IO_ADDR_R + 
					XARINA_NAND2_ECC_ERR_ADDR_BASE_REG + 
					XARINA_NAND2_ECC_ERROR_REG_OFFSET * i));
		e_bits = READ16(chip->IO_ADDR_R + XARINA_NAND2_ECC_ERR_BIT_BASE_REG + XARINA_NAND2_ECC_ERROR_REG_OFFSET * i);
		*(e_addr) ^= e_bits;
		error_corrected++;
	}

	return error_corrected;
}

#if defined(CONFIG_MTD_XARINA_NAND_DMAC)
static void xarina_nand2_dma_transfer(struct mtd_info *mtd, u_char *buf, int len, int direction)
{
	struct nand_chip *chip = mtd->priv;
	unsigned long data;
	int aligned;
	u_char *p_buf;

	/* AUTOCTL setting */
	data = XARINA_NAND2_NF_BIT_DM_TYPE | XARINA_NAND2_NF_BIT_AUTO_EN;
	if (direction == DMA_TO_DEVICE) {
		data |= XARINA_NAND2_NF_BIT_WXR;
	}
	writel(data, chip->IO_ADDR_W + XARINA_NAND2_NF_AUTOCTL);

	data = DMAC_RIN(DMAC_CONFIG);
	data |= DMAC_CONFIG_E;
	DMAC_ROUT(DMAC_CONFIG, data);

	p_buf = buf;
	/* address alignment check */
	aligned = ((unsigned long)buf & 0x3) ? 0 : 1;

	if (aligned == 0) {
		if (direction == DMA_TO_DEVICE) {
			memcpy(dmac_buf, buf, len);
		}
		p_buf = dmac_buf;
	}

	if (direction == DMA_FROM_DEVICE) {
		invalidate_dcache_range((unsigned long)p_buf, (unsigned long)(p_buf + len));
		DMAC_ROUT(DMAC_SRCADDR_0, (unsigned long)(chip->IO_ADDR_R + XARINA_NAND2_NF_FIFO));
		DMAC_ROUT(DMAC_DSTADDR_0, (unsigned long)p_buf);
		DMAC_ROUT(DMAC_LLI_0, DMAC_LLI_VAL);
		DMAC_ROUT(DMAC_DEF_LLI_0, DMAC_DEF_LLI_0_VAL);
		DMAC_ROUT(DMAC_DEF_CONTROL_0, DMAC_DEF_CONTROL_VAL_R);
		DMAC_ROUT(DMAC_DEF_CONFIG_0, DMAC_DEF_CONFIG_0_VAL_R);

	} else {
		flush_dcache_range((unsigned long)p_buf, (unsigned long)(p_buf + len));
#if defined(CONFIG_POSTEDWRITE_WORKAROUND)
		postedwrite_sync(POSTEDWRITE_DRAM);
#endif /* defined(CONFIG_POSTEDWRITE_WORKAROUND) */
		DMAC_ROUT(DMAC_SRCADDR_0, (unsigned long)p_buf);
		DMAC_ROUT(DMAC_DSTADDR_0, (unsigned long)(chip->IO_ADDR_R + XARINA_NAND2_NF_FIFO));
		DMAC_ROUT(DMAC_LLI_0, DMAC_LLI_VAL);
		DMAC_ROUT(DMAC_DEF_LLI_0, DMAC_DEF_LLI_0_VAL);
		DMAC_ROUT(DMAC_DEF_CONTROL_0, DMAC_DEF_CONTROL_VAL_W);
		DMAC_ROUT(DMAC_DEF_CONFIG_0, DMAC_DEF_CONFIG_0_VAL_W);

	}

	/* wait transfer end */
	while (!(readl(chip->IO_ADDR_W + XARINA_NAND2_NF_RIS) & 0x01));
	
#if defined(CONFIG_POSTEDWRITE_WORKAROUND)
	if (direction == DMA_FROM_DEVICE) {
		postedwrite_sync(POSTEDWRITE_A11DMAC_AHB1);
	}
#endif /* defined(CONFIG_POSTEDWRITE_WORKAROUND) */

	/* clear NFIF interrupt */
	writel(0x01, chip->IO_ADDR_W + XARINA_NAND2_NF_ICR);

	/* clear DMAC channel 0 interrupt */
	DMAC_ROUT(DMAC_INTTCCLR, INTTCCLEAR_0);

	if (aligned == 0) {
		if (direction == DMA_FROM_DEVICE) {
			memcpy(buf, p_buf, len);
		}
	}


	return;

}
#endif /* CONFIG_MTD_XARINA_NAND_DMAC */


/**
 * xarina_write_buf - write buffer to chip
 * @mtd:	MTD device structure
 * @buf:	data buffer
 * @len:	number of bytes to write
 *
 * Default write function for 8bit buswith
 */
static void xarina_write_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
	struct nand_chip *chip = mtd->priv;
	u_short task;
	int i;

	/* ecc clear */
	WRITE16(0xffff, chip->IO_ADDR_W + XARINA_NAND2_ECC_CLEAR_REG);

	/* set flash size = len */
	WRITE16((u_short)len, chip->IO_ADDR_W + XARINA_NAND2_FLASH_DATA_SIZE_REG);

	/* check control/status reg : FBSYn == 1 */
	while(chip->dev_ready(mtd) == XARINA_NAND2_DEVICE_BUSY);

	/* set task SendData/SendECC */
	task = XARINA_NAND2_BIT_TASK_SEND_DATA | xa_nand_info.ecc_task;
	xarina_nand2_set_task(mtd,task);

#if defined(CONFIG_MTD_XARINA_NAND_DMAC)
	if (len == XARINA_NAND2_ECC_SIZE){

		xarina_nand2_dma_transfer(mtd, (u_char *)buf, len, DMA_TO_DEVICE);

		/* check all tasks completed */
		while (1) {
			if (xarina_nand2_check_task(mtd, XARINA_NAND2_BIT_TASK_ALL) == 0)
				break;
		}

	} else {
#endif /* CONFIG_MTD_XARINA_NAND_DMAC */

		/* Write */
	for (i = 0; i < len / sizeof(u_short); i++) {
		/* NOT FIFO full check */
		while (1) {
			if ((READ16(chip->IO_ADDR_R + XARINA_NAND2_STATUS_REG)
					 & XARINA_NAND2_BIT_FIFO_FULL) == 0) {
				/* NOT FIFO FULL */
				break;
			}
		}	
		WRITE16(*((u_short *)buf + i), chip->IO_ADDR_W + XARINA_NAND2_DATA_REG);
	}

	/* wait for all the task finished */
	while (1) {
		if (xarina_nand2_check_task(mtd, XARINA_NAND2_BIT_TASK_ALL) == 0)
			break;
	}

#if defined(CONFIG_MTD_XARINA_NAND_DMAC)
	}
#endif /* CONFIG_MTD_XARINA_NAND_DMAC */

	return;
}

static int xarina_errstat(struct mtd_info *mtd, struct nand_chip *this, int state, int status, int page)
{
	struct erase_info einfo;

	switch (state) {
	case FL_WRITING:
		memset(&einfo, 0, sizeof(einfo));
		einfo.mtd = mtd;
		einfo.addr = (page << this->page_shift) & ~((1 << this->bbt_erase_shift) - 1);
		einfo.len = 1 << this->bbt_erase_shift;
		nand_erase_nand(mtd, &einfo, 1);
		mtd->block_markbad(mtd, page << this->page_shift);
		break;
	case FL_ERASING:
		mtd->block_markbad(mtd, page << this->page_shift);
		break;
	default:
		/* do nothing */
		break;
	}

	return status;
}

#endif /* !defined(CONFIG_NAND_SPL) */

/**
 * xarina_nand2_enable_hwecc
 * @mtd:    MTD device structure
 * mode:   Read / Write Mode
 *
 * Set Task for SendECC or ChkECC
 */
static void xarina_nand2_enable_hwecc(struct mtd_info *mtd, int mode)
{
	/* ecc task switch */
	
	switch (mode) {
	case NAND_ECC_READ:     /* read mode */

		/* Set Check ECC Task */
		xa_nand_info.ecc_task = XARINA_NAND2_BIT_TASK_CHK_ECC;

		break;

	case NAND_ECC_WRITE:    /* write mode */

		/* Set Send ECC Task */
		xa_nand_info.ecc_task = XARINA_NAND2_BIT_TASK_SEND_ECC;

		break;

	default:
		/* Clear ECC Task */
		xa_nand_info.ecc_task = 0;
		break;
	}
	return;
}

/**
 * xarina_nand2_write_protect
 *
 * set write_protect on or off
 * protect :  ON or OFF
*/	
void xarina_nand2_write_protect(struct mtd_info *mtd, int protect)
{
	register struct nand_chip *chip = mtd->priv;

	if (protect == XARINA_NAND2_WP_OFF) {
		/* write protect off*/
		WRITE16(READ16(chip->IO_ADDR_R + XARINA_NAND2_CONTROL_REG) | 
				 XARINA_NAND2_BIT_FWPN,
				 chip->IO_ADDR_W + XARINA_NAND2_CONTROL_REG);
	} else if (protect == XARINA_NAND2_WP_ON) {
	/* write protect on*/
		WRITE16(READ16(chip->IO_ADDR_R + XARINA_NAND2_CONTROL_REG) & 
				~XARINA_NAND2_BIT_FWPN, 
				chip->IO_ADDR_W + XARINA_NAND2_CONTROL_REG);
	} else {
#if !defined(CONFIG_NAND_SPL)
		printk("invalid value for write_protect\n");
#endif
	}


	return;
}
/**
 * xarina_dev_ready - chip status
 * @mtd:    MTD device structure
 *
 * Chip Status Return
*/
static int xarina_nand2_dev_ready(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	int chip_status;

	/* Ready Status Read */
	chip_status = READ16(chip->IO_ADDR_R + XARINA_NAND2_STATUS_REG) & XARINA_NAND2_BIT_FBSYN;

	if (chip_status) {
		return XARINA_NAND2_DEVICE_READY;	/* Ready */
	} else {
		return XARINA_NAND2_DEVICE_BUSY;	/* busy */
	}

	/* no pass */
	return XARINA_NAND2_DEVICE_READY;	/* Ready */

}


/*****************************************************************************/
/*  Chip Select for xarina nand                                              */
/*****************************************************************************/
static void xarina_select_chip(struct mtd_info *mtd, int chip)
{
	struct nand_chip *nand_chip = mtd->priv;
	unsigned long data;

	/* select chip enable/disable */

	data = READ16(nand_chip->IO_ADDR_R + XARINA_NAND2_CONTROL_REG) | 
			XARINA_NAND2_BIT_FCDISABLE;

	/* chip select */
	if( chip >= 0 && chip < XARINA_NAND2_MAX_NUM_CHIPS ){
		WRITE16(data & ~(XARINA_NAND2_BIT_FCE0N << chip), nand_chip->IO_ADDR_W + XARINA_NAND2_CONTROL_REG);
	} else {  /* chip deselect */
		/* disable all */
		WRITE16(data, nand_chip->IO_ADDR_W + XARINA_NAND2_CONTROL_REG);
		/* write protect on*/
		xarina_nand2_write_protect(mtd, XARINA_NAND2_WP_ON);
	}

	return;
}

/**
 * xarina_nand2_check_task
 *
 * check if task finished
 * task :  task_bits
*/
static int xarina_nand2_check_task(struct mtd_info *mtd, unsigned short task)
{
	int ret;
	register struct nand_chip *chip = mtd->priv;

	ret = READ16(chip->IO_ADDR_W + XARINA_NAND2_TASK_REG) & task;

	return ret;
}

/**
 * xarina_nand2_set_task
 *
 * write task function
 * task :  task_bits
*/
static void xarina_nand2_set_task(struct mtd_info *mtd, unsigned short task)
{

	register struct nand_chip *chip = mtd->priv;
	unsigned short tmp_task;

	tmp_task = READ16(chip->IO_ADDR_W + XARINA_NAND2_TASK_REG);

	tmp_task |= task;

	WRITE16(tmp_task, chip->IO_ADDR_W + XARINA_NAND2_TASK_REG);

	return;
}


/**
 * xarina_read_buf - read chip data into buffer
 * @mtd:	MTD device structure
 * @buf:	buffer to store date
 * @len:	number of bytes to read
 *
 * Default read function for 8bit buswith
 */
static void xarina_read_buf(struct mtd_info *mtd, u_char *buf, int len)
{
	struct nand_chip *chip = mtd->priv;
	u_short task;
	int i;

	/* ecc clear */
	WRITE16(0xffff, chip->IO_ADDR_W + XARINA_NAND2_ECC_CLEAR_REG);

	/* set flash size = len */
	WRITE16((u_short)len, chip->IO_ADDR_W + XARINA_NAND2_FLASH_DATA_SIZE_REG);

	/* set task RecvData/ChkECC */
	task = XARINA_NAND2_BIT_TASK_RECV_DATA | xa_nand_info.ecc_task;
	xarina_nand2_set_task(mtd,task);

#if defined(CONFIG_MTD_XARINA_NAND_DMAC)
	if (len == XARINA_NAND2_ECC_SIZE){

		xarina_nand2_dma_transfer(mtd, buf, len, DMA_FROM_DEVICE);

		/* check all tasks completed */
		while (1) {
			if (xarina_nand2_check_task(mtd, XARINA_NAND2_BIT_TASK_ALL) == 0)
				break;
		}

	} else {
#endif /* CONFIG_MTD_XARINA_NAND_DMAC */

	/* Read */
	for (i = 0; i < len / sizeof(u_short); i++) {
		/* NOT FIFO Empty Check */
		while (1) {
			if ((READ16(chip->IO_ADDR_R + XARINA_NAND2_STATUS_REG) & 
				XARINA_NAND2_BIT_FIFO_EMPTY) == 0) {
				/* not empty */
				break;
			}

		}

		*((u_short *)buf + i) = READ16(chip->IO_ADDR_R + XARINA_NAND2_DATA_REG);

	}

	/* check ecc calculation completed */
	while (1) {
		if (xarina_nand2_check_task(mtd, XARINA_NAND2_BIT_TASK_ALL) == 0)
			break;
	}

#if defined(CONFIG_MTD_XARINA_NAND_DMAC)
	}
#endif /* CONFIG_MTD_XARINA_NAND_DMAC */

	return;
}

/******************************************/
/* function for initialize                */
/******************************************/

/**
 * board_nand_info_set
 * @nand:   NAND device structure
 *
 * NAND Controller Initialize
 */
void board_nand2_info_set(struct nand_chip *nand)
{
	unsigned short tmp;
	int i;
	
	nand->IO_ADDR_R = nand->IO_ADDR_W = (void  __iomem *)XARINA_NANDIF_PHYS_BASE;
	
	nand->select_chip = xarina_select_chip;
	nand->dev_ready = xarina_nand2_dev_ready;

	nand->read_buf = xarina_read_buf;
#ifndef CONFIG_NAND_SPL
	nand->write_buf = xarina_write_buf;
#endif

#if !defined(CONFIG_NAND_SPL)

	nand->options = NAND_USE_FLASH_BBT | BBT_AUTO_REFRESH;
	nand->chip_delay = XARINA_NAND2_CHIP_DELAY;
	nand->cmd_ctrl = xarina_cmd_ctrl;
	nand->cmdfunc = xarina_nand2_command;
	nand->read_byte = xarina_read_byte;
	nand->read_word = xarina_read_word;
	nand->verify_buf = xarina_nand2_verify_buf;
	nand->bbt_td = &xarina_bbt_main_descr;
	nand->bbt_md = &xarina_bbt_mirror_descr;
	nand->badblock_pattern = &xarina_largepage_flashbased;
	nand->errstat = xarina_errstat;
	nand->ecc.mode = NAND_ECC_HW;
	nand->ecc.layout = &xarina_nand2_oob;
	nand->ecc.calculate = xarina_nand2_calculate_ecc;
	nand->ecc.correct = xarina_nand2_correct_data;
	nand->ecc.size = XARINA_NAND2_ECC_SIZE;
	nand->ecc.steps = XARINA_NAND2_ECC_STEPS;
	nand->ecc.bytes = XARINA_NAND2_ECC_BYTES;
	nand->ecc.total = XARINA_NAND2_ECC_TOTAL;
	nand->ecc.read_oob = xarina_nand2_read_oob_std;
	nand->ecc.write_oob = xarina_nand2_write_oob_std;
	nand->ecc.read_page_raw = xarina_nand2_read_page_raw;
	nand->ecc.write_page_raw = xarina_nand2_write_page_raw;
	nand->ecc.read_page = xarina_nand2_read_page_hwecc;
	nand->ecc.write_page = xarina_nand2_write_page_hwecc;
#endif /* !defined(CONFIG_NAND_SPL) */

	nand->ecc.hwctl = xarina_nand2_enable_hwecc;

	/* task all clear */
	tmp = (READ16(nand->IO_ADDR_W + XARINA_NAND2_TASK_REG) | XARINA_NAND2_BIT_TASK_CLR);
	WRITE16(tmp, nand->IO_ADDR_W + XARINA_NAND2_TASK_REG);

	/* set read_id count 0 */
	xa_nand_info.status_id_read_cnt = 0;
	/* initialize read id/status buffer */
	for (i = 0; i < 6; i++) {	/* NAND SPL can not use memset() */
		xa_nand_info.read_byte_reg[i] = 0;
	}
	/* set ecc_task 0 */
	xa_nand_info.ecc_task = 0;


	/* Set Timing Registers */
	WRITE16( XARINA_NAND2_PARAM_CMD_RATE,  nand->IO_ADDR_W + XARINA_NAND2_COMMAND_RATE_REG);
	WRITE16( XARINA_NAND2_PARAM_RW_RATE,   nand->IO_ADDR_W + XARINA_NAND2_RW_RATE_REG);
	WRITE16( XARINA_NAND2_PARAM_BUSYWAIT1, nand->IO_ADDR_W + XARINA_NAND2_BUSY_WAIT1_REG);
	WRITE16( XARINA_NAND2_PARAM_BUSYWAIT2, nand->IO_ADDR_W + XARINA_NAND2_BUSY_WAIT2_REG);

#if defined(CONFIG_MTD_XARINA_NAND_DMAC)
	/* Enable WAIT_ACT for NAND DMAC */
	WRITE16(READ16(nand->IO_ADDR_W + XARINA_NAND2_MODE_REG) | XARINA_NAND2_BIT_WAIT_ACT, 
			nand->IO_ADDR_W + XARINA_NAND2_MODE_REG);
	/* enable interrupt */
	writel(0x01, nand->IO_ADDR_W + XARINA_NAND2_NF_IEN);
#endif


	/* make FIFO empty */
	while ((READ16(nand->IO_ADDR_R + XARINA_NAND2_STATUS_REG) & XARINA_NAND2_BIT_FIFO_EMPTY) == 0) {
		tmp = READ16(nand->IO_ADDR_R + XARINA_NAND2_DATA_REG);
	}

	return;
}




#endif /* (CONFIG_COMMANDS & CFG_CMD_NAND) && !defined(CFG_NAND_LEGACY) */


