/*
 * (C) Copyright 2007, 2011 Sony Corporation.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <common.h>
#include <command.h>
#include <asm/arch/xarina.h>
#include <nand.h>
#include "cmd_update.h"
#include "hw_reg_ctrl.h"


#define MAIN_TABLE		(0)
#define MIRR_TABLE		(1)
#define INVALID			(0xff)
#define MAX_LOGICAL_BLOCK	(10)
#define MARK_LEN		(4)
#define TABLE_NUM		(2)	/* main table and mirror table */
#define MAX_TBL_SET		(2)
#define TBL_SET_X		(0)
#define TBL_SET_Y		(1)
#define OOB_FREE_MAX 		(32)
#define NAND_PAGE_COUNT		(64)
#define UPDATE_ERROR		(-1)
#define UPDATE_SUCCESS		(0)
#define UPDATE_RETRY_COUNT	(3)



/* #define DEBUG */

#ifdef DEBUG
#define DBG(format, arg...)	printk(KERN_ALERT "UPDATE: " format, ## arg)
#else
#define DBG(format, arg...)
#endif


typedef struct s_xa_oob {
	unsigned char mark[4];
	unsigned char loblock;
	unsigned char contflag;
	unsigned char version;	
	unsigned char reserved;	
} xa_oob;

typedef struct s_lblk_tbl_set {
	unsigned char lblk[TABLE_NUM][MAX_LOGICAL_BLOCK];
	unsigned char version;
} lblk_tbl_set;

static const unsigned char mark_spl[] = {'S', 'P', 'L', ' ' };
static const unsigned char mark_uboot[] = {'u', 'b', 'o', 't' };
static lblk_tbl_set xarina_lblk[MAX_TBL_SET];
static lblk_tbl_set write_lblk;
static int nand_ecc_steps;
static int nand_oob_free_size;
static int nand_block_size;
static int nand_page_size;
static int nand_oob_count;

static void set_oob_data(unsigned char *dst, xa_oob *oob)
{
	int i;
	
	for (i = 0; i < nand_oob_count; i++) {
		memcpy(dst + i * sizeof(xa_oob), oob, sizeof(xa_oob));
	}	
		
	return;
}


static unsigned long get_next_block(unsigned long start_block, 
				unsigned long end_block, unsigned long cur_block)
{
	unsigned long ret;
	
	if (end_block < start_block) {
		return start_block;
	}
	
	if (cur_block < start_block || end_block < cur_block) {
		return start_block;
	}
	
	ret = cur_block + 1;
	
	if (end_block < ret) {
		ret = start_block;
	}
	
	return ret;
}

/**
 * cmp_mark
 *
 * commpare two marks in oob
 * @src0: 	address to mark0
 * @src1:	address to mark1
 * @retval:	0	two marks are different
 * @retval:	1	two marks are same 
 */
static int cmp_mark(unsigned char *src0, const unsigned char *src1)
{
	int i;
	int ret = 1;
	
	for(i = 0; i < MARK_LEN; i++) {
		if(src0[i] != src1[i]) {
			ret = 0;
		}
	}
	return ret;
}
/**
 * init_lblk_tbl_set
 *
 * fill logical block table with 0xff.
 * @table: 	address to logical block table
 * @return:	none
 */
static void init_lblk_tbl_set(lblk_tbl_set *table)
{
	int i ,j;
	
	for(i = 0; i < TABLE_NUM; i++) {
		for (j = 0; j < MAX_LOGICAL_BLOCK; j++) {
			table->lblk[i][j] = INVALID;
		}
	}
	table->version = INVALID;
}

/**
 * set_blk_num
 *
 * set physical block number to logical block table.
 * @table: 	address to logical block table
 * @lblk_num: 	logical block number
 * @pblk_num: 	physical block number
 * @return:	none
 */
static void set_blk_num(lblk_tbl_set *table, unsigned char lblk_num, unsigned char pblk_num)
{
	if (table->lblk[MAIN_TABLE][lblk_num] == INVALID) {
		table->lblk[MAIN_TABLE][lblk_num] = pblk_num;
	} else if (table->lblk[MIRR_TABLE][lblk_num] == INVALID) {
		table->lblk[MIRR_TABLE][lblk_num] = pblk_num;
	} else {
		/* do nothing */
	}
}


/**
 * cmp_version
 *
 * compare two numbers.
 * notice: 0x00 is bigger than 0xff.
 * @ver0: 	version number to compare	
 * @ver1: 	version number to compare	
 * @retval:	0	ver1 is smaller than ver0
 * @retval:	1	ver1 is bigger than ver0
 */
static int cmp_version(unsigned char ver0, unsigned char ver1)
{
	if (ver0 < ver1) {
		return (0 < ((ver1 - ver0 + 0x01) & 0xff));
	} else {
		return (1 > ((ver0 - ver1 + 0x01) & 0xff)); 
	}
}

static int cmp_oob(unsigned char *oob)
{
	xa_oob* oob_data = (xa_oob *)oob;
	int ret = 0;
	int i;

	for (i = 1; i < nand_oob_count; i++) {
		ret += memcmp(oob_data[0].mark, oob_data[i].mark, 4);
	}

	if (ret != 0) {
		DBG("cmp_oob() error\n");
	}

	return ret;
}

static int create_lblk_table(struct mtd_info *mtd, int offset, 
			    int block_len, unsigned char *ver_cnt,
			    const unsigned char *mark)
{

	int i, j;
	unsigned char oob_buf[OOB_FREE_MAX];
	struct mtd_oob_ops ops;
	xa_oob *oob_data;
	int ret = 1;

	/* initialize tables */
	init_lblk_tbl_set(&(xarina_lblk[TBL_SET_X]));
	init_lblk_tbl_set(&(xarina_lblk[TBL_SET_Y]));
	*ver_cnt = 0;
	
	printf("scanning NAND ");
	DBG("PBN:LBN:VER (valid block list)\n");
	
	for (i = 0; i < block_len; i++) {
		if (i % 10 == 0) {
			printf(".");
		}

		ops.ooblen = sizeof(xa_oob) * nand_oob_count;
		ops.oobbuf = oob_buf;
		ops.ooboffs = 0;
		ops.datbuf = NULL;
		ops.mode = MTD_OOB_PLACE;
		ops.len = nand_page_size;

		for (j = 0; j < NAND_PAGE_COUNT; j++) {
			mtd->read_oob(mtd, (i + offset) * nand_block_size, &ops);
			if (!cmp_oob(oob_buf)) {
				break;
			}
		}
		if (j == NAND_PAGE_COUNT) {
			continue;
		}
		

		oob_data = (xa_oob *)(oob_buf);
		if (cmp_mark(oob_data->mark, mark) && 
		    (oob_data->loblock < MAX_LOGICAL_BLOCK)) {
			switch (*ver_cnt) {
			case 0:
				xarina_lblk[TBL_SET_X].version = oob_data->version;
				set_blk_num(&(xarina_lblk[TBL_SET_X]), 
					oob_data->loblock, i + offset);
				(*ver_cnt)++;
				break;
			case 1:
				if (xarina_lblk[TBL_SET_X].version == oob_data->version) {
					set_blk_num(&(xarina_lblk[TBL_SET_X]), 
						oob_data->loblock, i + offset);
				} else {
					xarina_lblk[TBL_SET_Y].version = oob_data->version;
					set_blk_num(&(xarina_lblk[TBL_SET_Y]), 
						oob_data->loblock, i + offset);
					(*ver_cnt)++;
				}
				break;
			default:
				if (xarina_lblk[TBL_SET_X].version == oob_data->version) {
					set_blk_num(&(xarina_lblk[TBL_SET_X]), 
						oob_data->loblock, i + offset);
				} else if (xarina_lblk[TBL_SET_Y].version == oob_data->version){
					set_blk_num(&(xarina_lblk[TBL_SET_Y]), 
					oob_data->loblock, i + offset);
				} else {
					(*ver_cnt)++;
				}
				break;
			
			}
			
			DBG("%02x : %02x : %02x \n", 
				i + offset, oob_data->loblock, oob_data->version);
		}
	}

	DBG("version count: %02x\n", *ver_cnt);

	for(j = 0; j < *ver_cnt; j++) { 
		if (i >= MAX_TBL_SET) {
			break;
		}
		DBG("version %02x ,main table ----------------------\n", xarina_lblk[j].version);
		for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
			DBG("%02x:%02x\n", i, xarina_lblk[j].lblk[MAIN_TABLE][i]); 
		}
		DBG("version %02x ,mirror table ----------------------\n", xarina_lblk[j].version);
		for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
			DBG("%02x:%02x\n", i, xarina_lblk[j].lblk[MIRR_TABLE][i]); 
		}
	}
	
	if ((1 <= *ver_cnt) && (*ver_cnt <= 2)) {
		ret = 0;
	}
	
	printf(" done!\n");
	
	return ret;
}

static unsigned long get_start_block(int tbl_set, unsigned long start_block,
					unsigned long end_block)
{
	int i, j;
	unsigned long ret = 0;
	
	if (tbl_set >= MAX_TBL_SET) {
		DBG("ERROR:get_start_block()\n");
		return INVALID;
	}
	
	for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
		for (j = 0; j < TABLE_NUM; j++) {
			if ((xarina_lblk[tbl_set].lblk[j][i] != INVALID) && 
			    (xarina_lblk[tbl_set].lblk[j][i] > ret)) {
			    ret = xarina_lblk[tbl_set].lblk[j][i];
			}
		}
	}
	
	DBG("get_start_block() : %lx\n", get_next_block(start_block, end_block, ret));
	return get_next_block(start_block, end_block, ret);
	
}

static int check_blk_regster(lblk_tbl_set *table, unsigned char pblk)
{
	int i, j;
	
	if (table == NULL) {
		DBG("ERROR:check_blk_regster\n");
		return 0;
	}
	
	for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
		for (j = 0; j < TABLE_NUM; j++) {
			if (table->lblk[j][i] == pblk) {
				return 1;
			}
		}
	}	
	return 0;
}

static int get_write_block(nand_info_t *nand, unsigned char base_blk,
			 unsigned char start_blk, unsigned char end_blk,
			 lblk_tbl_set *excl_table)
{
	
	unsigned char block = base_blk;

	while (1) {
		if (!nand_block_isbad(nand, block * nand_block_size) &&
			!check_blk_regster(&write_lblk, block) &&
			!check_blk_regster(excl_table, block)) {
			return block;
		} else {
			block = get_next_block(start_blk, end_blk, block);
			if (block == base_blk) {
				DBG("ERROR: can not get write block");
				return -1;
			}
		}
	}
	
	return -1;
}

static void erase_block_from_table(nand_info_t *nand, lblk_tbl_set *table)
{
	int i ,j;
	
	printf("erasing version 0x%02x ... ", table->version);
	
	for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
		for (j = 0; j < TABLE_NUM; j++) {
			if (table->lblk[j][i] != INVALID) {
				nand_erase(nand, 
					table->lblk[j][i] * nand_block_size, 
					nand_block_size);
				DBG("erase : %x\n",
					 table->lblk[j][i] * nand_block_size);
			}
		}
	}
	printf("done!\n");

}

static void leave_block_from_table(nand_info_t *nand, lblk_tbl_set *table, 
				unsigned char start_blk, unsigned char end_blk)
{
	int i;
	
	for (i = start_blk; i <= end_blk; i++) {
		if (!check_blk_regster(table, i)) {
			nand_erase(nand, i * nand_block_size,
				nand_block_size);
		}
	}
}

static int update_core(nand_info_t *nand, int tbl_set, const unsigned char *mark,
				 unsigned char base_blk, unsigned char start_blk, 
				 unsigned char end_blk, u32 addr, int size)
{
	unsigned char write_ver;
	unsigned char tmp_oob[OOB_FREE_MAX];
	unsigned char *write_addr;
	xa_oob oob_data;
	int i, j;
	int p_total = 0;
	int p_total_tmp = 0;
	int write_page;
	int b_len;
	int p_len;
	int write_blk;
	int ret;
	struct mtd_oob_ops ops;
	struct nand_chip *chip = nand->priv;

	b_len = ((size - 1) >> chip->phys_erase_shift) + 1;/* block length */
	p_len = ((size - 1) >> chip->page_shift) + 1;	/* page length	 */
	

	DBG("base_blk = %x\n", base_blk);
	DBG("start_blk = %x\n", start_blk);
	DBG("end_blk = %x\n", end_blk);

	/* set version to be written */
	if (tbl_set == INVALID || tbl_set >= MAX_TBL_SET) {
		write_ver = 0;
		tbl_set = TBL_SET_X;
	} else {
		write_ver = (xarina_lblk[tbl_set].version + 1) & 0xff;
	}

	write_blk = (int)base_blk;

	/* tmp_oob init */
	memset(tmp_oob, 0xff , OOB_FREE_MAX);

	/* initialize oob data */
	memcpy(oob_data.mark, mark, 4);
	oob_data.loblock = 0;
	oob_data.contflag = 0xff;
	oob_data.version = write_ver;
	oob_data.reserved = 0xff;
	
	chip->select_chip(nand, 0);
	
	printf("update version 0x%02x ", write_ver);
	
	
	for(i = 0; i < b_len * 2; i++) {
		oob_data.loblock = i / 2;
		oob_data.contflag = 0xff;
		set_oob_data(tmp_oob, &oob_data);
		write_blk = get_write_block(nand, write_blk, 
					start_blk, end_blk, &xarina_lblk[tbl_set]);

		if (write_blk < 0) {
			printf("... failure!\n");
			return UPDATE_ERROR;
		}
		ret = nand_erase(nand, write_blk * nand_block_size, 
				nand_block_size);

		if (ret) {
			return UPDATE_ERROR;	
		}
		
		p_total_tmp = p_total;
		printf(".");
		DBG("write:0x%08x\n",
			write_blk * NAND_PAGE_COUNT * nand_page_size);

		for (j = 0; j < NAND_PAGE_COUNT; j++) {
			write_page = j + write_blk * NAND_PAGE_COUNT;
			write_addr = (unsigned char *)(addr + 
					i / 2 * nand_block_size + 
					j * nand_page_size);
			DBG("write addr:0x%08x\n", write_addr);
			ops.ooblen = nand_oob_free_size;
			ops.oobbuf = tmp_oob;
			ops.ooboffs = 0;
			ops.datbuf = write_addr;
			ops.mode = MTD_OOB_PLACE;
			ops.len = nand_page_size;
				
			if (p_total_tmp == (p_len - 1)) {
				
				oob_data.contflag = 0x00;
				set_oob_data(tmp_oob, &oob_data);
				ret = nand->write_oob(nand, 
					write_page * nand_page_size, &ops);
				DBG("last page = 0x%08x\n", 
					write_page * nand_page_size); 
				if (ret) {
					return UPDATE_ERROR;	
				}
				break;
			}
			DBG("write page:0x%08x\n", write_page * nand_page_size);
			ret = nand->write_oob(nand, 
				write_page * nand_page_size, &ops);
			if (ret) {
				return UPDATE_ERROR;	
			}
			p_total_tmp++;
		}

		if (i % 2) {
			p_total += NAND_PAGE_COUNT;
		}
		write_lblk.lblk[i % 2][i / 2] = write_blk;
		write_blk = get_next_block(start_blk, end_blk, write_blk);
	}
	printf(" done!\n");

	return UPDATE_SUCCESS;	
	
}

static void nand_info_get(nand_info_t *nand)
{
	struct nand_chip *chip;

	chip = nand->priv;
	nand_ecc_steps = chip->ecc.steps;
	nand_oob_free_size = nand->oobsize - chip->ecc.layout->eccbytes;
	nand_block_size = 1 << chip->phys_erase_shift;
	nand_page_size = 1 << chip->page_shift;
	if (xarina_hw_nand_cont_type_get () == XARINA_NAND_CONTROLLER1) {
		nand_oob_count = 4;
	} else {
		nand_oob_count = 2;
	}

	DBG("nand_ecc_steps = %d\n", nand_ecc_steps);
	DBG("nand_oob_free_size = %d\n", nand_oob_free_size);
	DBG("nand_block_size = %d\n", nand_block_size);
	DBG("nand_page_size = %d\n", nand_page_size);
	DBG("NAND_PAGE_COUNT = %d\n", NAND_PAGE_COUNT);


}


static int update_loader(u32 addr, int size, 
			unsigned long start_ofs, unsigned long end_ofs,
			const unsigned char *mark)
{
	unsigned char ver_cnt = 0;
	nand_info_t *nand;
	int tbl_set;
	unsigned long base_block;
	unsigned long start_block;
	unsigned long end_block;
	int i;
	int ret;

	/* the following commands operate on the current device */
	if (nand_curr_device < 0 || nand_curr_device >= CONFIG_SYS_MAX_NAND_DEVICE ||
	    !nand_info[nand_curr_device].name) {
		puts("\nno devices available\n");
		return 1;
	}
	
	nand = &nand_info[nand_curr_device];

	nand_info_get(nand);

	start_block = start_ofs / nand_block_size;
	end_block = end_ofs / nand_block_size;

	create_lblk_table(nand, start_block, 
			end_block - start_block + 1, &ver_cnt, mark);
	printf("found %d version ", ver_cnt);
	if ((1 <= ver_cnt) && (ver_cnt <= 2)) {
		for (i = 0; i < ver_cnt; i++) {
			if (i != 0) {
				printf(", ") ;
			}
			printf("0x%02x", xarina_lblk[i].version);
		}
	}
	printf("\n");
	
	init_lblk_tbl_set(&write_lblk);

	switch (ver_cnt) {
	case 0:
		tbl_set = INVALID;
		base_block = start_block;
		ret = update_core(nand, tbl_set, mark, base_block,
				 start_block, end_block, addr, size);
		break;
	case 1:
		tbl_set = TBL_SET_X;
		base_block = get_start_block(tbl_set, start_block, end_block);
		ret = update_core(nand, tbl_set, mark, base_block,
				start_block, end_block, addr, size);
		break;
	case 2:
		tbl_set = cmp_version(xarina_lblk[TBL_SET_X].version, 
		      xarina_lblk[TBL_SET_Y].version) ? TBL_SET_Y : TBL_SET_X;
		/* erase old version */
		erase_block_from_table(nand, &xarina_lblk[tbl_set ^0x01]);
		init_lblk_tbl_set(&xarina_lblk[tbl_set ^0x01]);
		base_block = get_start_block(tbl_set, start_block, end_block);
		ret = update_core(nand, tbl_set, mark, base_block, 
			start_block, end_block, addr, size);

		break;
	default :
		tbl_set = cmp_version(xarina_lblk[TBL_SET_X].version, 
		      xarina_lblk[TBL_SET_Y].version) ? TBL_SET_Y : TBL_SET_X;
		printf("current data in NAND is invalid\n");
		printf("erase ... ");
		leave_block_from_table(nand, &xarina_lblk[tbl_set], 
				start_block, end_block);
		printf("done!\n");
		init_lblk_tbl_set(&xarina_lblk[tbl_set ^0x01]);
		base_block = get_start_block(tbl_set, start_block, end_block);
		ret = update_core(nand, tbl_set, mark, base_block, 
			start_block, end_block, addr, size);
		
		break;
	}

	return ret;
}



int do_update(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret = UPDATE_ERROR;
	int i;
	u32 addr;
	u32 size;
	u32 start_ofs;
	u32 end_ofs;
	
	if (argc != 4) {
		cmd_usage(cmdtp);
		return 0;			
	}
	addr = (u32)simple_strtoul(argv[2], NULL, 16);
	size = (u32)simple_strtoul(argv[3], NULL, 16);


	if (strcmp(argv[1], "u-boot") == 0 ) {
		if (xarina_hw_nand_cont_type_get() == XARINA_NAND_CONTROLLER1) {
			start_ofs = CONFIG_XARINA_NAND1_PART_UBOOT_START;
			end_ofs = CONFIG_XARINA_NAND1_PART_UBOOT_END;
		} else {
			start_ofs = CONFIG_XARINA_NAND2_PART_UBOOT_START;
			end_ofs = CONFIG_XARINA_NAND2_PART_UBOOT_END;
		}
		for (i = 0; i < UPDATE_RETRY_COUNT; i++) {
			ret = update_loader(addr, size, start_ofs, end_ofs, mark_uboot);
			if (ret == UPDATE_SUCCESS) {
				break;
			} else {
				printf("\n%s update is failure. retry...\n", argv[1]);
			}
		}

	} else if (strcmp(argv[1], "spl") == 0 ) {
		if (xarina_hw_nand_cont_type_get() == XARINA_NAND_CONTROLLER1) {
			start_ofs = CONFIG_XARINA_NAND1_PART_SPL_START;
			end_ofs = CONFIG_XARINA_NAND1_PART_SPL_END;
		} else {
			start_ofs = CONFIG_XARINA_NAND2_PART_SPL_START;
			end_ofs = CONFIG_XARINA_NAND2_PART_SPL_END;
		}
		for (i = 0; i < UPDATE_RETRY_COUNT; i++) {
			ret = update_loader(addr, size, start_ofs, end_ofs, mark_spl);
			if (ret == UPDATE_SUCCESS) {
				break;
			} else {
				printf("\n%s update is failure. retry...\n", argv[1]);
			}
		}
	}

	if (ret == UPDATE_SUCCESS) {
		printf("\n%s update is done!\n", argv[1]);
	} else {
		printf("\n%s update is failure.\n", argv[1]);
	}
	
	return ret;
}

U_BOOT_CMD(
	update, 4, 0,	do_update,
	"update u-boot or spl",
	"<target (u-boot | spl)> < addr > < size >\n"
	"    - update 'target' from 'addr' to NAND\n"
	"    - 'target' size is 'size'\n"
);



static int test_update(unsigned char *addr, int offset, int size, unsigned char version, 
	unsigned char interval, int rev, unsigned char *mark)
{
	nand_info_t *nand;
	struct nand_chip *chip;
	xa_oob oob_data;
	unsigned char tmp_oob[OOB_FREE_MAX];
	int b_len;
	int p_len;
	int p_total = 0;
	int i, j;
	int write_page;
	unsigned char *write_addr;
	int sign;
	struct mtd_oob_ops ops;


	/* the following commands operate on the current device */
	if (nand_curr_device < 0 || nand_curr_device >= CONFIG_SYS_MAX_NAND_DEVICE ||
	    !nand_info[nand_curr_device].name) {
		puts("\nno devices available\n");
		return 1;
	}
	nand = &nand_info[nand_curr_device];
	chip = nand->priv;

	b_len = ((size - 1) >> chip->phys_erase_shift) + 1;/* block length */
	p_len = ((size - 1) >> chip->page_shift) + 1;	/* page length	 */
	
	printf("b_len = %d\n", b_len); 
	printf("p_len = %d\n", p_len); 

	nand_info_get(nand);

	memset(tmp_oob, 0xff , OOB_FREE_MAX);

	/* initialize oob data */
	memcpy(oob_data.mark, mark, 4);
	if (oob_data.mark[3] == '\0')
	{
		oob_data.mark[3] = ' ';
	}
	oob_data.loblock = 0;
	oob_data.contflag = 0xff;
	oob_data.version = version;
	oob_data.reserved = 0xff;
	offset = offset >> chip->page_shift;
	chip->select_chip(nand, 0);
	
	sign = (rev != 0) ? -1 : 1;
	
	for(i = 0; i < b_len; i++) {
		oob_data.loblock = i;
		set_oob_data(tmp_oob, &oob_data);
		for (j = 0; j < NAND_PAGE_COUNT; j++) {
			write_page = offset + j + sign * ((int)interval + 1) * i * NAND_PAGE_COUNT;
			write_addr = addr + i * nand_block_size + j * nand_page_size;
			if (j == 0) {
				printf("write:0x%08x\n", write_page * nand_page_size);
			}
			ops.ooblen = nand_oob_free_size;
			ops.oobbuf = tmp_oob;
			ops.ooboffs = 0;
			ops.datbuf = write_addr;
			ops.mode = MTD_OOB_PLACE;
			ops.len = nand_page_size;
				
			if (p_total == (p_len - 1)) {
				
				oob_data.contflag = 0x00;
				set_oob_data(tmp_oob, &oob_data);
				/*
				write_page_oob(nand, chip, write_page,
					write_addr, tmp_oob, OOB_FREE_MAX);
				*/
				nand->write_oob(nand, write_page * nand_page_size, &ops);
				printf("last page = %d\n", write_page); 
				break;
			}
			nand->write_oob(nand, write_page * nand_page_size, &ops);

			p_total++;
		}
	}
	
	return 0;
}

int do_test_update(int argc, char * const argv[])
{
	unsigned char *addr;
	unsigned long offset;
	unsigned long size;
	unsigned long version;
	unsigned long interval;
	unsigned long reverse = 0;
	
	if ((argc < 7) || (8 < argc)) {
		printf("usage: %s address offset size(byte) version interval mark [r]\n", argv[0]);
		return 0;
	}
	
	printf("update start\n");
	addr = (unsigned char*)simple_strtoul(argv[1], NULL, 16);
	offset = simple_strtoul(argv[2], NULL, 16);
	size = simple_strtoul(argv[3], NULL, 16);
	version = simple_strtoul(argv[4], NULL, 16);
	interval = simple_strtoul(argv[5], NULL, 16);

	if(*(argv[7]) == 'r') {
		reverse = 1;
	}
	
	printf("address = %08lx\n", (unsigned long)addr);
	printf("offset = %08lx\n", offset);
	printf("size = %08lx\n", size);
	printf("version = %08lx\n", version);
	printf("interval = %08lx\n", interval);
	if (reverse) {
		printf("reverse\n");
	}
	
	test_update(addr, (int)offset, (int)size, 
		(unsigned char)version, (unsigned char)interval, (int)reverse, (unsigned char *)argv[6]);
	
	return 0;
}






