/*
 * (C) Copyright 2012 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <version.h>
#include <common.h>
#include <command.h>
#include "web_update.h"
#include "cmd_menu.h"

extern char cmd_buff[CONFIG_SYS_CBSIZE];
extern int ubifs_get_file_size(char *filename, int *size);
extern int ubifs_load(char *filename, u32 addr, u32 size);
extern int ubifs_ls(char *filename);

/*
 * mount_ubifs_vol
 * @part: the partition name to be mounted
 * @vol: the volumn name to be mounted
 * return 0 if success
 *
 * This function mounts the specified 'vol' on 'part'.
 */
static int mount_ubifs_vol(char *part, char *vol)
{
	int ret;
	char part_name[256];
	char vol_name[256];

	ret = getenv_f(part, part_name, 256);
	if (ret == -1) {
		printf("error: env \"%s\" is not found.\n", part);
		return 1;
	}

	ret = getenv_f(vol, vol_name, 256);

	if (ret == -1) {
		printf("error: env \"%s\" is not found.\n", vol);
	return 1;
	}

	sprintf(cmd_buff, "ubi part %s", part_name);

	ret = run_command(cmd_buff, 0);
	if (ret < 0) {
		printf("ubi attach failure\n");
		return 1;
	}

	sprintf(cmd_buff, "ubifsmount %s", vol_name);
	ret = run_command(cmd_buff, 0);

	if (ret < 0) {
		printf("ubifs mount failure\n");
		return 1;
	}

	return ret;
}

/*
 * is_update_file_exist
 * @update_file_name: the file name of the update image
 * return 0: file does not exist
 *        1: file exists
 *
 * This function checks whether the update image file exists.
 * Assumes the volume containing the file has already mounted.
 */
int is_update_file_exist(char *update_file_name)
{
	int ret;

	ret = ubifs_ls(update_file_name);

	return ret==0?1:0;
}

/*
 * update_nand_spl
 * @update_file_name: the file name of the update image
 * return 0 if success
 *
 * This function loads the spl image file to memory then update nand spl.
 * Assumes the volume containing the update file has already mounted.
 */
int update_nand_spl(char *update_file_name)
{
	int ret;
	int file_size;

	ret = ubifs_get_file_size(update_file_name, &file_size);

	if (ret) {
		printf("get nand spl file size error\n");
		return ret;
	}

	ret = ubifs_load(update_file_name, CONFIG_SYS_LOAD_ADDR, file_size);

	if (ret) {
		printf("load nand spl file error\n");
		return ret;
	}

	sprintf(cmd_buff, "update spl %x %x", CONFIG_SYS_LOAD_ADDR, file_size);

	printf("cmd: %s\n", cmd_buff);

	ret = run_command(cmd_buff, 0);

	return ret;
}

/*
 * update_uboot
 * @update_file_name: the file name of the update image
 * return 0 if success
 *
 * This function loads the uboot image file to memory then update uboot.
 * Assumes the volume containing the update file has already mounted.
*/
int update_uboot(char *update_file_name)
{
	int ret;
	int file_size;

	ret = ubifs_get_file_size(update_file_name, &file_size);

	if (ret) {
		printf("get uboot file size error\n");
		return ret;
	}

	ret = ubifs_load(update_file_name, CONFIG_SYS_LOAD_ADDR, file_size);

	if (ret) {
		printf("load uboot file error\n");
		return ret;
	}

	sprintf(cmd_buff, "update u-boot %x %x", CONFIG_SYS_LOAD_ADDR, file_size);

	printf("cmd: %s\n", cmd_buff);

	ret = run_command(cmd_buff, 0);

	return ret;
}

/*
 * update_mtd_volume
 * @update_file_name: the file name of the update image
 * @part: the destination partition to be updated
 * @vol: the destination volume to be updated
 * return 0 if success
 *
 * This function loads the update image file to memory from src_part, src_vol.
 * Then update the part, vol.
 * Assumes the volume containing the update file has already mounted.
*/
int update_mtd_volume(char *update_file_name, char *part, char *vol)
{
	int ret;
	int file_size;
	char part_name[256];
	char vol_name[256];

	ret = ubifs_get_file_size(update_file_name, &file_size);

	if (ret) {
		printf("get file size error\n");
		return ret;
	}

	ret = getenv_f(part, part_name, 256);
	if (ret == -1) {
		printf("error: env \"%s\" is not found.\n", CONFIG_XARINA_ENV_UBI_PART);
		return 1;
	}

	printf("%s:part_name = %s\n", __func__, part_name);

	ret = getenv_f(vol, vol_name, 256);

	if (ret == -1) {
		printf("error: env \"%s\" is not found.\n", CONFIG_XARINA_ENV_UBI_VOLUME);
		return 1;
	}

	printf("%s:vol_name = %s\n", __func__, vol_name);

	ret = ubifs_load(update_file_name, CONFIG_SYS_LOAD_ADDR, file_size);

	if (ret) {
		printf("load file error\n");
		return ret;
	}

	sprintf(cmd_buff, "nand erase %s", part_name);
	printf("%s\n",cmd_buff);
	ret = run_command(cmd_buff, 0);
	if (ret < 0) {
		printf("nand erase failure\n");
		return 1;
	}

	sprintf(cmd_buff, "ubi part %s", part_name);
	printf("%s\n", cmd_buff);
	ret = run_command(cmd_buff, 0);
	if (ret < 0) {
		printf("ubi attach failure\n");
		return 1;
	}

	sprintf(cmd_buff, "ubi create %s", vol_name);
	printf("%s\n", cmd_buff);
	ret = run_command(cmd_buff, 0);
	if (ret < 0) {
		printf("ubi volume create failure\n");
		return 1;
	}

	sprintf(cmd_buff, "ubi write %x %s %x",CONFIG_SYS_LOAD_ADDR, vol_name, file_size);
	printf("%s\n", cmd_buff);
	ret = run_command(cmd_buff, 0);
	if (ret < 0) {
		printf("ubi write failure\n");
		return 1;
	}
/*
	sprintf(cmd_buff, "ubifsmount %s", vol_name);
	ret = run_command(cmd_buff, 0);
	if (ret < 0) {
		printf("ubifs mount failure\n");
		return 1;
	}
*/
	printf("\nubifs update is done!\n\n");

	return ret;
}

/*
 * get_web_update_indicator
 * return the indicator value, or 0 if failed to get value or
 * value gotten in invalid.
 *
 * This function gets the value of uboot env 'web_update_indicator.
 */
int get_web_update_indicator(void)
{
	char str[256];
	int ret;

	ret = getenv_f(CONFIG_XARINA_WEB_UPDATE_INDICATOR, str, 256);

	printf("CONFIG_XARINA_WEB_UPDATE_INDICATOR is %c \n",str[0]);
	if (ret == -1) {
		printf("web update indicator was not found, using the default value 0\n");
		return 0;
	}

	if (str[0] >= '0' && str[0] <= '9')
		return str[0] - '0';
	if (str[0] >= 'a' && str[0] <= 'f')
		return str[0] - 'a' + 0x0A;
	if (str[0] >= 'A' && str[0] <= 'F')
		return str[0] - 'A' + 0x0A;

	return 0;
}

/*
 * get_lastsensor_cmdnum
 * return the last menu preset command number after web update,
 * or value 0 if failed to get value or value gotten in invalid.
 *
 * This function gets last menu preset command number.
 */
int  get_lastsensor_cmdnum(void)
{
	char env_men_preset_str[64];
	char env_men_input_str[64];

	int i,j;

	char *chr_env_men_preset_str = NULL;
	char *chr_env_men_input_str = NULL;

	char *video_in_sensor_name = "Video" ; 
	char *imx_140la_sensor_name = "IMX140LA";

	sprintf(env_men_input_str,"menu_input_%02d_%02d_str",1,1);
	chr_env_men_input_str = getenv(env_men_input_str);

	if(chr_env_men_input_str == NULL){
		printf("can not get menu input str,using the default value 0 \n");
		return 0;
	}
	
	if(strstr(chr_env_men_input_str,video_in_sensor_name) != NULL){
		printf("Find the Video in target \n");
		return 1;
	}

	if(strstr(chr_env_men_input_str,imx_140la_sensor_name) != NULL){
		printf("Find the IMX140LA target \n");
		for(j = 0; j < MENU_ITEM_NUM_MAX; j++)
		{
			sprintf(env_men_preset_str,"menu_preset_%02d_str",j+1);
			chr_env_men_preset_str = getenv(env_men_preset_str);
			if(chr_env_men_preset_str == NULL ){
				return j;
			}
		}
	}

	for(i = 0; i < MENU_ITEM_NUM_MAX; i++)
	{
		sprintf(env_men_preset_str,"menu_preset_%02d_str",i+1);

		chr_env_men_preset_str = getenv(env_men_preset_str);
		if(chr_env_men_preset_str == NULL ){
			printf("can not get menu preset str, using the default value 0 \n");
			return 0;
		}

		if(strstr(chr_env_men_input_str,chr_env_men_preset_str) != NULL ){
			printf("Find the target: %s, and will execute the command:menu_preset_%02d_cmd \n",
				chr_env_men_preset_str,i+1);
			return i + 1;
		}
	}

	return 0;
}

/* execute_update_corefw_cmd
*  @selecnum: the value of last menu preset command number
*
*  This function update CoreFw environment values.
*  For example, if Video in, execute the command 
*  "set menu_preset_01_cmd setenv filesize\;ubifsload 61000000 update/vin.ush\;script 61000000 \$(filesize)"
*/ 
void execute_update_corefw_cmd(int selecnum)
{
	char env[64];
	char cmd_buff_temp[256];
	int ret;

	sprintf(env,"menu_preset_%02d_cmd",selecnum);
	ret = getenv_f(env,cmd_buff_temp,256);

	if(ret == -1){
		printf(" env %s is not defined. web update  execute_update_corefw_cmd  failed \n",
			env);
		return;
	}
	printf("web update  execute_update_corefw_cmd is %s /n",cmd_buff_temp);
	run_command(cmd_buff_temp,0);
}

/*
 * set_web_update_indicator
 * @indicator: the value of indicator to be written to env
 *
 * This function write the 'indicator' to uboot env "web_update_indictor"
 */
void set_web_update_indicator(int indicator)
{
	sprintf(cmd_buff, "setenv %s %x", CONFIG_XARINA_WEB_UPDATE_INDICATOR,
						indicator);
	run_command(cmd_buff, 0);
	run_command("saveenv", 0);
}

int xarina_do_web_update(void)
{
	int ret;
	char ch;
	int indicator;

	indicator = get_web_update_indicator();

	/* if indicator has no flag to update, return directly */
	if (indicator == 0)
		return 0;

	ret = mount_ubifs_vol(CONFIG_XARINA_ENV_STREAM_UBI_PART,
		CONFIG_XARINA_ENV_STREAM_UBI_VOLUME);


	if (ret != 0) {
		goto xarina_do_web_update_out;
	}

	if (!is_update_file_exist(CONFIG_XARINA_WEB_UPDATE_FOLDER)) {
	printf("No update folder found!\n");
	goto xarina_do_web_update_out;
	}

#ifdef WEBUPDATE_PROMPT_USER
	printf("Found web update folder, do you want to update now? [y/n]?");
	ch = getc();
	printf("%c\n", ch);
	if (ch != 'y' && ch != 'Y') {
	printf("Skip web update\n");
	goto xarina_do_web_update_out;
	}
#else
	printf("Found web update folder... \n");
#endif

	if ((indicator & WEB_UPDATE_MASK_SPL) &&
		is_update_file_exist(CONFIG_XARINA_WEB_UPDATE_NANDSPL_FILE_NAME)) {
#ifdef WEBUPDATE_PROMPT_USER
		printf("Found nand spl update file, update? [y/n]?");
		ch = getc();
		printf("%c\n", ch);
		if (ch == 'y' || ch == 'Y') {
			ret = update_nand_spl(CONFIG_XARINA_WEB_UPDATE_NANDSPL_FILE_NAME);
			indicator &= ~WEB_UPDATE_MASK_SPL;
		}
#else
		printf("Update nand spl file... \n" );
		ret = update_nand_spl(CONFIG_XARINA_WEB_UPDATE_NANDSPL_FILE_NAME);
		indicator &= ~WEB_UPDATE_MASK_SPL;
#endif

	}

	if ((indicator & WEB_UPDATE_MASK_UBOOT) &&
		is_update_file_exist(CONFIG_XARINA_WEB_UPDATE_UBOOT_FILE_NAME)) {
#ifdef WEBUPDATE_PROMPT_USER
		printf("Found U-boot update file, update? [y/n]?");
		ch = getc();
		printf("%c\n", ch);
		if (ch == 'y' || ch == 'Y') {
			ret = update_uboot(CONFIG_XARINA_WEB_UPDATE_UBOOT_FILE_NAME);
			indicator &= ~WEB_UPDATE_MASK_UBOOT;
		}
#else
		printf("Update U-boot update file... \n");
		ret = update_uboot(CONFIG_XARINA_WEB_UPDATE_UBOOT_FILE_NAME);
		indicator &= ~WEB_UPDATE_MASK_UBOOT;
#endif
	}

	if ((indicator & WEB_UPDATE_MASK_COREFW) &&
			is_update_file_exist(CONFIG_XARINA_WEB_UPDATE_FW_FILE_NAME)) {
#ifdef WEBUPDATE_PROMPT_USER
		printf("Found CoreFW update file, update? [y/n]?");
		ch = getc();
		printf("%c\n", ch);
		if (ch == 'y' || ch == 'Y') {
			ret = update_mtd_volume(CONFIG_XARINA_WEB_UPDATE_FW_FILE_NAME,
						CONFIG_XARINA_ENV_UBI_PART,
						CONFIG_XARINA_ENV_UBI_VOLUME);
			indicator &= ~WEB_UPDATE_MASK_COREFW;
		}
#else
		printf("Update CoreFW update file... \n");
		ret = update_mtd_volume(CONFIG_XARINA_WEB_UPDATE_FW_FILE_NAME,
					CONFIG_XARINA_ENV_UBI_PART,
					CONFIG_XARINA_ENV_UBI_VOLUME);
		indicator &= ~WEB_UPDATE_MASK_COREFW;
#endif
	}

	ret = mount_ubifs_vol(CONFIG_XARINA_ENV_STREAM_UBI_PART,
		CONFIG_XARINA_ENV_STREAM_UBI_VOLUME);

	if (ret != 0)
		goto xarina_do_web_update_out;


	if ((indicator & WEB_UPDATE_MASK_ROOTFS) &&
		is_update_file_exist(CONFIG_XARINA_WEB_UPDATE_ROOTFS_FILE_NAME)) {
#ifdef WEBUPDATE_PROMPT_USER
		printf("Found Root File System update file, update? [y/n]?");
		ch = getc();
		printf("%c\n", ch);
		if (ch == 'y' || ch == 'Y') {
			ret = update_mtd_volume(CONFIG_XARINA_WEB_UPDATE_ROOTFS_FILE_NAME,
						CONFIG_XARINA_ENV_ROOTFS_UBI_PART,
						CONFIG_XARINA_ENV_ROOTFS_UBI_VOLUME);
			indicator &= ~WEB_UPDATE_MASK_ROOTFS;
		}
#else
		printf("Udpate Root File System update file... \n");
		ret = update_mtd_volume(CONFIG_XARINA_WEB_UPDATE_ROOTFS_FILE_NAME,
					CONFIG_XARINA_ENV_ROOTFS_UBI_PART,
					CONFIG_XARINA_ENV_ROOTFS_UBI_VOLUME);
		indicator &= ~WEB_UPDATE_MASK_ROOTFS;
#endif
	}

	/* update the indicator */
	indicator = 0;
	set_web_update_indicator(indicator);

	ret = mount_ubifs_vol(CONFIG_XARINA_ENV_UBI_PART,
				CONFIG_XARINA_ENV_UBI_VOLUME);
	if(ret != 0)
		goto xarina_do_web_update_out;	

	int lastsensor_cmdnum = get_lastsensor_cmdnum();
	if(lastsensor_cmdnum == 0)
		goto xarina_do_web_update_out;

	execute_update_corefw_cmd(lastsensor_cmdnum);

	run_command("reset", 0);

xarina_do_web_update_out:

	ret = mount_ubifs_vol(CONFIG_XARINA_ENV_UBI_PART,
		CONFIG_XARINA_ENV_UBI_VOLUME);

	return ret;
}
