#include <config.h>
#include <common.h>
#include <command.h>
#include <nand.h>
#include <asm/arch/xarina_nand.h>
#include "cmd_update.h"

#define XARINA_NAND_OOB_MAX 	(32)

#define MAIN_TABLE		(0)
#define MIRR_TABLE		(1)
#define INVALID			(0xff)
#define MAX_LOGICAL_BLOCK	(10)
#define MARK_LEN		(4)
#define TABLE_NUM		(2)	/* main table and mirror table */
#define MAX_TBL_SET		(2)
#define TBL_SET_X		(0)
#define TBL_SET_Y		(1)
/* #define DEBUG */

#ifdef DEBUG
#define DBG(format, arg...)	printk(KERN_ALERT "UPDATE: " format, ## arg)
#else
#define DBG(format, arg...)
#endif


typedef struct {
	unsigned char mark[4];
	unsigned char loblock;
	unsigned char contflag;
	unsigned char version;	
	unsigned char reserved;	
} xa_oob;

typedef struct {
	unsigned char lblk[TABLE_NUM][MAX_LOGICAL_BLOCK];
	unsigned char version;
} lblk_tbl;

static unsigned char mark_spl[] = {'S', 'P', 'L', ' ' };
static unsigned char mark_uboot[] = {'u', 'b', 'o', 't' };
static lblk_tbl xarina_lblk[MAX_TBL_SET];
static lblk_tbl write_lblk;



static void set_oob_data(unsigned char *dst, xa_oob *oob)
{
	int i;
	
	for (i = 0; i < 4; i++) {
		memcpy(dst + i * sizeof(xa_oob), oob, sizeof(xa_oob));
	}	
		
	return;
}


static unsigned long get_next_block(unsigned long start_block, 
				unsigned long end_block, unsigned long cur_block)
{
	unsigned long ret;
	
	if (end_block < start_block) {
		return start_block;
	}
	
	if (cur_block < start_block || end_block < cur_block) {
		return start_block;
	}
	
	ret = cur_block + 1;
	
	if (end_block < ret) {
		ret = start_block;
	}
	
	return ret;
}

/**
 * cmp_mark
 *
 * commpare two marks in oob
 * @src0: 	address to mark0
 * @src1:	address to mark1
 * @retval:	0	two marks are different
 * @retval:	1	two marks are same 
 */
static int cmp_mark(unsigned char *src0, const unsigned char *src1)
{
	int i;
	int ret = 1;
	
	for(i = 0; i < MARK_LEN; i++) {
		if(src0[i] != src1[i]) {
			ret = 0;
		}
	}
	return ret;
}
/**
 * init_lblk_tbl
 *
 * fill logical block table with 0xff.
 * @table: 	address to logical block table
 * @return:	none
 */
static void init_lblk_tbl(lblk_tbl *table)
{
	int i ,j;
	
	for(i = 0; i < TABLE_NUM; i++) {
		for (j = 0; j < MAX_LOGICAL_BLOCK; j++) {
			table->lblk[i][j] = INVALID;
		}
	}
	table->version = INVALID;
}

/**
 * set_blk_num
 *
 * set physical block number to logical block table.
 * @table: 	address to logical block table
 * @lblk_num: 	logical block number
 * @pblk_num: 	physical block number
 * @return:	none
 */
static void set_blk_num(lblk_tbl *table, unsigned char lblk_num, unsigned char pblk_num)
{
	if (table->lblk[MAIN_TABLE][lblk_num] == INVALID) {
		table->lblk[MAIN_TABLE][lblk_num] = pblk_num;
	} else if (table->lblk[MIRR_TABLE][lblk_num] == INVALID) {
		table->lblk[MIRR_TABLE][lblk_num] = pblk_num;
	} else {
		/* do nothing */
	}
}


/**
 * cmp_version
 *
 * compare two numbers.
 * notice: 0x00 is bigger than 0xff.
 * @ver0: 	version number to compare	
 * @ver1: 	version number to compare	
 * @retval:	0	ver1 is smaller than ver0
 * @retval:	1	ver1 is bigger than ver0
 */
static int cmp_version(unsigned char ver0, unsigned char ver1)
{
	if (ver0 < ver1) {
		return (0 < ((ver1 - ver0 + 0x01) & 0xff));
	} else {
		return (1 > ((ver0 - ver1 + 0x01) & 0xff)); 
	}
}

static int cmp_oob(unsigned char *oob)
{
	xa_oob* oob_data[CONFIG_SYS_NAND_ECCSTEPS];
	int ret = 0;
	int i;
	
	for (i = 0; i < CONFIG_SYS_NAND_ECCSTEPS; i++) {
		oob_data[i] = (xa_oob *)(oob + sizeof(xa_oob) * i);
	}
	
	for(i = 1; i < CONFIG_SYS_NAND_ECCSTEPS; i++) {
		ret += memcmp(oob_data[0]->mark, oob_data[i]->mark, 4);							
	}
	
	if (ret != 0) {
		DBG("cmp_oob() error\n");
	}

	return ret;
}

static int create_lbk_table(struct mtd_info *mtd, int offset, 
			    int block_len, unsigned char *ver_cnt,
			    unsigned char *mark)
{

	int i, j;
	unsigned char oob_buf[CONFIG_XARINA_NAND_OOB_SIZE];
	struct mtd_oob_ops ops;
	xa_oob *oob_data;
	int ret = 1;

	/* initialize tables */
	init_lblk_tbl(&(xarina_lblk[TBL_SET_X]));
	init_lblk_tbl(&(xarina_lblk[TBL_SET_Y]));
	*ver_cnt = 0;
	
	printf("scanning NAND ");
	DBG("PBN:LBN:VER (valid block list)\n");
	
	for (i = 0; i < block_len; i++) {
		if (i % 10 == 0) {
			printf(".");
		}
		if (nand_block_isbad(mtd, (i + offset) * CONFIG_SYS_NAND_BLOCK_SIZE)) { 
			continue;
		}
		ops.ooblen = CONFIG_XARINA_NAND_OOB_SIZE;
		ops.oobbuf = oob_buf;
		ops.ooboffs = 0;
		ops.datbuf = NULL;
		ops.mode = MTD_OOB_PLACE;
		ops.len = CONFIG_SYS_NAND_PAGE_SIZE;

		for (j = 0; j < CONFIG_SYS_NAND_PAGE_COUNT; j++) {
			mtd->read_oob(mtd, (i + offset) * CONFIG_SYS_NAND_BLOCK_SIZE, &ops);
			if (!cmp_oob(oob_buf)) {
				break;
			}
		}
		if (j == CONFIG_SYS_NAND_PAGE_COUNT) {
			continue;
		}
		

		oob_data = (xa_oob *)(oob_buf);
		if (cmp_mark(oob_data->mark, mark) && 
		    (oob_data->loblock < MAX_LOGICAL_BLOCK)) {
			switch (*ver_cnt) {
			case 0:
				xarina_lblk[TBL_SET_X].version = oob_data->version;
				set_blk_num(&(xarina_lblk[TBL_SET_X]), 
					oob_data->loblock, i + offset);
				(*ver_cnt)++;
				break;
			case 1:
				if (xarina_lblk[TBL_SET_X].version == oob_data->version) {
					set_blk_num(&(xarina_lblk[TBL_SET_X]), 
						oob_data->loblock, i + offset);
				} else {
					xarina_lblk[TBL_SET_Y].version = oob_data->version;
					set_blk_num(&(xarina_lblk[TBL_SET_Y]), 
						oob_data->loblock, i + offset);
					(*ver_cnt)++;
				}
				break;
			default:
				if (xarina_lblk[TBL_SET_X].version == oob_data->version) {
					set_blk_num(&(xarina_lblk[TBL_SET_X]), 
						oob_data->loblock, i + offset);
				} else if (xarina_lblk[TBL_SET_Y].version == oob_data->version){
					set_blk_num(&(xarina_lblk[TBL_SET_Y]), 
					oob_data->loblock, i + offset);
				} else {
					(*ver_cnt)++;
				}
				break;
			
			}
			
			DBG("%02x : %02x : %02x \n", 
				i + offset, oob_data->loblock, oob_data->version);
		}
	}

	DBG("version count: %02x\n", *ver_cnt);

	for(j = 0; j < *ver_cnt; j++) { 
		if (i >= MAX_TBL_SET) {
			break;
		}
		DBG("version %02x ,main table ----------------------\n", xarina_lblk[j].version);
		for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
			DBG("%02x:%02x\n", i, xarina_lblk[j].lblk[MAIN_TABLE][i]); 
		}
		DBG("version %02x ,mirror table ----------------------\n", xarina_lblk[j].version);
		for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
			DBG("%02x:%02x\n", i, xarina_lblk[j].lblk[MIRR_TABLE][i]); 
		}
	}
	
	if ((1 <= *ver_cnt) && (*ver_cnt <= 2)) {
		ret = 0;
	}
	
	printf(" done!\n");
	
	return ret;
}

static unsigned long get_start_block(int tbl_set, unsigned long start_block,
					unsigned long end_block)
{
	int i, j;
	unsigned long ret = 0;
	
	if (tbl_set >= MAX_TBL_SET) {
		DBG("ERROR:get_start_block()\n");
		return INVALID;
	}
	
	for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
		for (j = 0; j < TABLE_NUM; j++) {
			if ((xarina_lblk[tbl_set].lblk[j][i] != INVALID) && 
			    (xarina_lblk[tbl_set].lblk[j][i] > ret)) {
			    ret = xarina_lblk[tbl_set].lblk[j][i];
			}
		}
	}
	
	DBG("get_start_block() : %lx\n", get_next_block(start_block, end_block, ret));
	return get_next_block(start_block, end_block, ret);
	
}

static int check_blk_regster(lblk_tbl *table, unsigned char pblk)
{
	int i, j;
	
	if (table == NULL) {
		DBG("ERROR:check_blk_regster\n");
		return 0;
	}
	
	for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
		for (j = 0; j < TABLE_NUM; j++) {
			if (table->lblk[j][i] == pblk) {
				return 1;
			}
		}
	}	
	return 0;
}

static int get_write_block(nand_info_t *nand, unsigned char base_blk,
			 unsigned char start_blk, unsigned char end_blk,
			 lblk_tbl *excl_table)
{
	
	unsigned char block = base_blk;

	while (1) {
		if (!nand_block_isbad(nand, block * CONFIG_SYS_NAND_BLOCK_SIZE) &&
			!check_blk_regster(&write_lblk, block) &&
			!check_blk_regster(excl_table, block)) {
			return block;
		} else {
			block = get_next_block(start_blk, end_blk, block);
			if (block == base_blk) {
				DBG("ERROR: can not get write block");
				return -1;
			}
		}
	}
	
	return -1;
}

static void erase_block_from_table(nand_info_t *nand, lblk_tbl *table)
{
	int i ,j;
	
	printf("erasing version 0x%02x ... ", table->version);
	
	for (i = 0; i < MAX_LOGICAL_BLOCK; i++) {
		for (j = 0; j < TABLE_NUM; j++) {
			if (table->lblk[j][i] != INVALID) {
				nand_erase(nand, 
					table->lblk[j][i] * CONFIG_SYS_NAND_BLOCK_SIZE, 
					CONFIG_SYS_NAND_BLOCK_SIZE);
				DBG("erase : %x\n",
					 table->lblk[j][i] * CONFIG_SYS_NAND_BLOCK_SIZE);
			}
		}
	}
	printf("done!\n");

}

static void leave_block_from_table(nand_info_t *nand, lblk_tbl *table, 
				unsigned char start_blk, unsigned char end_blk)
{
	int i;
	
	for (i = start_blk; i <= end_blk; i++) {
		if (!check_blk_regster(table, i)) {
			nand_erase(nand, i * CONFIG_SYS_NAND_BLOCK_SIZE,
				CONFIG_SYS_NAND_BLOCK_SIZE);
		}
	}
}

static int update_core(nand_info_t *nand, int tbl_set, unsigned char *mark,
				 unsigned char base_blk, unsigned char start_blk, 
				 unsigned char end_blk, u32 addr, int size)
{
	unsigned char write_ver;
	unsigned char tmp_oob[XARINA_NAND_OOB_MAX];
	unsigned char *write_addr;
	xa_oob oob_data;
	int i, j;
	int p_total = 0;
	int p_total_tmp = 0;
	int write_page;
	int b_len = (size - 1) / CONFIG_SYS_NAND_BLOCK_SIZE + 1;/* block length */
	int p_len = (size - 1) / CONFIG_SYS_NAND_PAGE_SIZE + 1;	/* page length	 */
	int write_blk;
	struct mtd_oob_ops ops;
	struct nand_chip *chip = nand->priv;

	DBG("base_blk = %x\n", base_blk);
	DBG("start_blk = %x\n", start_blk);
	DBG("end_blk = %x\n", end_blk);

	/* set version to be written */
	if (tbl_set == INVALID || tbl_set >= MAX_TBL_SET) {
		write_ver = 0;
	} else {
		write_ver = (xarina_lblk[tbl_set].version + 1) & 0xff;
	}

	write_blk = (int)base_blk;
	
	/* initialize oob data */
	memcpy(oob_data.mark, mark, 4);
	oob_data.loblock = 0;
	oob_data.contflag = 0xff;
	oob_data.version = write_ver;
	oob_data.reserved = 0xff;
	
	chip->select_chip(nand, 0);
	
	printf("update version 0x%02x ", write_ver);
	
	
	for(i = 0; i < b_len * 2; i++) {
		oob_data.loblock = i / 2;
		set_oob_data(tmp_oob, &oob_data);
		write_blk = get_write_block(nand, write_blk, 
					start_blk, end_blk, &xarina_lblk[tbl_set]);

		if (write_blk < 0) {
			printf("... failure!\n");
			return 1;
		}
		nand_erase(nand, write_blk * CONFIG_SYS_NAND_BLOCK_SIZE, 
				CONFIG_SYS_NAND_BLOCK_SIZE);
		
		p_total_tmp = p_total;
		printf(".");
		DBG("write:0x%08x\n",
			write_blk * CONFIG_SYS_NAND_PAGE_COUNT * CONFIG_SYS_NAND_PAGE_SIZE);

		for (j = 0; j < CONFIG_SYS_NAND_PAGE_COUNT; j++) {
			write_page = j + write_blk * CONFIG_SYS_NAND_PAGE_COUNT;
			write_addr = (unsigned char *)(addr + 
					i / 2 * CONFIG_SYS_NAND_BLOCK_SIZE + 
					j * CONFIG_SYS_NAND_PAGE_SIZE);
			ops.ooblen = CONFIG_XARINA_NAND_OOB_SIZE;
			ops.oobbuf = tmp_oob;
			ops.ooboffs = 0;
			ops.datbuf = write_addr;
			ops.mode = MTD_OOB_PLACE;
			ops.len = CONFIG_SYS_NAND_PAGE_SIZE;
				
			if (p_total_tmp == (p_len - 1)) {
				
				oob_data.contflag = 0x00;
				set_oob_data(tmp_oob, &oob_data);
				/*
				write_page_oob(nand, chip, write_page,
					write_addr, tmp_oob, XARINA_NAND_OOB_MAX);
				*/
				nand->write_oob(nand, 
					write_page * CONFIG_SYS_NAND_PAGE_SIZE, &ops);
				DBG("last page = 0x%08x\n", 
					write_page * CONFIG_SYS_NAND_PAGE_SIZE); 
				break;
			}
			nand->write_oob(nand, 
				write_page * CONFIG_SYS_NAND_PAGE_SIZE, &ops);
			p_total_tmp++;
		}
		if (i % 2) {
			p_total += CONFIG_SYS_NAND_PAGE_COUNT;
		}
		write_lblk.lblk[i % 2][j] = write_blk;
		write_blk = get_next_block(start_blk, end_blk, write_blk);
	}
	printf(" done!\n");

	return 0;	
	
}

static int update_loader(u32 addr, int size, 
			unsigned long start_ofs, unsigned long end_ofs,
			unsigned char *mark)
{

#if !defined(CONFIG_XARINA_NAND_NO_USE)
	unsigned char ver_cnt = 0;
	nand_info_t *nand;
	int tbl_set;
	unsigned long base_block;
	unsigned long start_block = start_ofs / CONFIG_SYS_NAND_BLOCK_SIZE;
	unsigned long end_block = end_ofs / CONFIG_SYS_NAND_BLOCK_SIZE;
	int i;

	/* the following commands operate on the current device */
	if (nand_curr_device < 0 || nand_curr_device >= CONFIG_SYS_MAX_NAND_DEVICE ||
	    !nand_info[nand_curr_device].name) {
		puts("\nno devices available\n");
		return 1;
	}
	
	nand = &nand_info[nand_curr_device];

	create_lbk_table(nand, start_block, 
			end_block - start_block + 1, &ver_cnt, mark);
	printf("found %d version ", ver_cnt);
	if ((1 <= ver_cnt) && (ver_cnt <= 2)) {
		for (i = 0; i < ver_cnt; i++) {
			if (i != 0) {
				printf(", ") ;
			}
			printf("0x%02x", xarina_lblk[i].version);
		}
	}
	printf("\n");
	
	init_lblk_tbl(&write_lblk);

	switch (ver_cnt) {
	case 0:
		tbl_set = INVALID;
		base_block = start_block;
		update_core(nand, tbl_set, mark, base_block,
				 start_block, end_block, addr, size);
		break;
	case 1:
		tbl_set = TBL_SET_X;
		base_block = get_start_block(tbl_set, start_block, end_block);
		update_core(nand, tbl_set, mark, base_block,
				start_block, end_block, addr, size);
		break;
	case 2:
		tbl_set = cmp_version(xarina_lblk[TBL_SET_X].version, 
		      xarina_lblk[TBL_SET_Y].version) ? TBL_SET_Y : TBL_SET_X;
		/* erase old version */
		erase_block_from_table(nand, &xarina_lblk[tbl_set ^0x01]);
		init_lblk_tbl(&xarina_lblk[tbl_set ^0x01]);
		base_block = get_start_block(tbl_set, start_block, end_block);
		update_core(nand, tbl_set, mark, base_block, 
			start_block, end_block, addr, size);

		break;
	default :
		tbl_set = cmp_version(xarina_lblk[TBL_SET_X].version, 
		      xarina_lblk[TBL_SET_Y].version) ? TBL_SET_Y : TBL_SET_X;
		printf("current data in NAND is invalid\n");
		printf("erase ... ");
		leave_block_from_table(nand, &xarina_lblk[tbl_set], 
				start_block, end_block);
		printf("done!\n");
		init_lblk_tbl(&xarina_lblk[tbl_set ^0x01]);
		base_block = get_start_block(tbl_set, start_block, end_block);
		update_core(nand, tbl_set, mark, base_block, 
			start_block, end_block, addr, size);
		
		break;
	}
#endif
	return 1;
}



int do_update(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret = 0;
	u32 addr;
	u32 size;
	
	if (argc != 4) {
		cmd_usage(cmdtp);
		return 0;			
	}
	addr = (u32)simple_strtoul(argv[2], NULL, 16);
	size = (u32)simple_strtoul(argv[3], NULL, 16);


	if (strcmp(argv[1], "u-boot") == 0 ) {
		ret = update_loader(addr, size, 
				CONFIG_XARINA_NAND_PART_UBOOT_START,
				CONFIG_XARINA_NAND_PART_UBOOT_END,
				mark_uboot);
	} else if (strcmp(argv[1], "spl") == 0 ) {
		ret = update_loader(addr, size, 
				CONFIG_XARINA_NAND_PART_SPL_START,
				CONFIG_XARINA_NAND_PART_SPL_END,
				mark_spl);
	}

	if (ret) {
		printf("%s update is done!\n", argv[1]);
	}
	
	return 0;
}

U_BOOT_CMD(
	update, 4, 0,	do_update,
	"update u-boot or spl",
	"<target (u-boot | spl)> < addr > < size >\n"
	"    - update 'target' from 'addr' to NAND\n"
	"    - 'target' size is 'size'\n"
);


#if defined(CONFIG_FOR_DEVELOPMENT)

static int test_update(unsigned char *addr, int offset, int size, unsigned char version, 
	unsigned char interval, int rev)
{
#if !defined(CONFIG_XARINA_NAND_NO_USE)
	nand_info_t *nand;
	struct nand_chip *chip;
	xa_oob oob_data;
	unsigned char tmp_oob[XARINA_NAND_OOB_MAX];
	int b_len = (size - 1) / CONFIG_SYS_NAND_BLOCK_SIZE + 1;
	int p_len = (size - 1) / CONFIG_SYS_NAND_PAGE_SIZE + 1;
	int p_total = 0;
	int i, j;
	int write_page;
	unsigned char *write_addr;
	int sign;
	struct mtd_oob_ops ops;
	
	
	printf("b_len = %d\n", b_len); 
	printf("p_len = %d\n", p_len); 

	/* the following commands operate on the current device */
	if (nand_curr_device < 0 || nand_curr_device >= CONFIG_SYS_MAX_NAND_DEVICE ||
	    !nand_info[nand_curr_device].name) {
		puts("\nno devices available\n");
		return 1;
	}
	nand = &nand_info[nand_curr_device];
	chip = nand->priv;
	
	/* initialize oob data */
	memcpy(oob_data.mark, mark_uboot, 4);
	oob_data.loblock = 0;
	oob_data.contflag = 0xff;
	oob_data.version = version;
	oob_data.reserved = 0xff;
	
	chip->select_chip(nand, 0);
	
	sign = (rev != 0) ? -1 : 1;
	
	for(i = 0; i < b_len; i++) {
		oob_data.loblock = i;
		set_oob_data(tmp_oob, &oob_data);
		for (j = 0; j < CONFIG_SYS_NAND_PAGE_COUNT; j++) {
			write_page = offset + j + sign * ((int)interval + 1) * i * CONFIG_SYS_NAND_PAGE_COUNT;
			write_addr = addr + i * CONFIG_SYS_NAND_BLOCK_SIZE + j * CONFIG_SYS_NAND_PAGE_SIZE;
			if (j == 0) {
				printf("write:0x%08x\n", write_page * CONFIG_SYS_NAND_PAGE_SIZE);
			}
			ops.ooblen = CONFIG_XARINA_NAND_OOB_SIZE;
			ops.oobbuf = tmp_oob;
			ops.ooboffs = 0;
			ops.datbuf = write_addr;
			ops.mode = MTD_OOB_PLACE;
			ops.len = CONFIG_SYS_NAND_PAGE_SIZE;
				
			if (p_total == (p_len - 1)) {
				
				oob_data.contflag = 0x00;
				set_oob_data(tmp_oob, &oob_data);
				/*
				write_page_oob(nand, chip, write_page,
					write_addr, tmp_oob, XARINA_NAND_OOB_MAX);
				*/
				nand->write_oob(nand, write_page * CONFIG_SYS_NAND_PAGE_SIZE, &ops);
				printf("last page = %d\n", write_page); 
				break;
			}
			nand->write_oob(nand, write_page * CONFIG_SYS_NAND_PAGE_SIZE, &ops);
			//write_page_oob(nand, chip,  write_page,
			//	write_addr, tmp_oob, XARINA_NAND_OOB_MAX);
			p_total++;
		}
	}
#endif
	return 0;
}

int do_test_update(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	unsigned char *addr;
	unsigned long offset;
	unsigned long size;
	unsigned long version;
	unsigned long interval;
	unsigned long reverse = 0;
	
	if ((argc < 6) || (7 < argc)) {
		printf("usage: %s address offset size(byte) version interval [r]\n", argv[0]);
		return 0;
	}
	
	printf("update start\n");
	addr = (unsigned char*)simple_strtoul(argv[1], NULL, 16);
	offset = simple_strtoul(argv[2], NULL, 16);
	size = simple_strtoul(argv[3], NULL, 16);
	version = simple_strtoul(argv[4], NULL, 16);
	interval = simple_strtoul(argv[5], NULL, 16);

	if(*(argv[6]) == 'r') {
		reverse = 1;
	}
	
	printf("address = %08lx\n", (unsigned long)addr);
	printf("offset = %08lx\n", offset);
	printf("size = %08lx\n", size);
	printf("version = %08lx\n", version);
	printf("interval = %08lx\n", interval);
	if (reverse) {
		printf("reverse\n");
	}
	
	test_update(addr, (int)(offset / CONFIG_SYS_NAND_PAGE_SIZE), (int)size, 
		(unsigned char)version, (unsigned char)interval, (int)reverse);
	
	return 0;
}


U_BOOT_CMD(
	testupdate, 7, 0,	do_test_update,
	"write U-Boot binary to NAND (for Development)",
	"write U-Boot binary to NAND (for Development)"
);

#endif 




