/*
 * (C) Copyright 2007, 2011 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>

#include <malloc.h>
#include <asm/io.h>
#include <linux/compiler.h>
#include "smac_mii.h"
#ifdef CONFIG_XARINA
#include <asm/arch/hw_reg.h>
#endif /* CONFIG_XARINA */

#if defined(CONFIG_SMAC) && defined(CONFIG_NET_MULTI)

/********************************/
/* low level access             */
/********************************/
static unsigned long g_smac_regbase;
/* register read for smac */
static unsigned long smac_readreg(unsigned long adr)
{
	adr = g_smac_regbase + adr;
	return readl((void *) adr);
}

/* register write for smac */
static void smac_writereg(unsigned long adr, unsigned long data)
{
	adr = g_smac_regbase + adr;
	writel(data, (void *) adr);

	return;
}

/* phy register access busy wait */
static void smac_wait_phy(void)
{
	while (smac_readreg(SMAC_PHY_ADDR) & SMAC_PA_M_GB) {
		/* wait phy work done */
	}

	return;
}

/* smac_read_desc - read data from descriptor */
static uint32_t smac_read_desc(void* addr)
{
	return *(uint32_t*)addr;
}


/* smac_write_desc - write data to descriptor */
static void smac_write_desc(uint32_t val, void* addr)
{
	*(uint32_t*)addr = val;

#ifdef CONFIG_XARINA
	/* work-around for posted write */
	{
		volatile uint32_t dummy; 

		barrier();
		dummy = smac_read_desc(addr);
	}
#endif
}


#ifdef CONFIG_XARINA
/* register write for sysreg1 */
static void smac_write_clock_reg(unsigned long clock)
{
	unsigned long data;
	/* GbE clock disable */
	data = SYSREG1_RIN(SYSREG1_CLK_EN);
	data &= ~SYSREG1_GBE_CKON;
	data &= ~SYSREG1_GBEEXT_EN;
	SYSREG1_ROUT(SYSREG1_CLK_EN, data);
	/* clock select */	
	data = SYSREG1_RIN(SYSREG1_CLK_CTL);
	data = (data & ~SYSREG1_GBE_CKSEL_MASK) | (clock << SYSREG1_GBE_CKSEL_SHIFT);
	SYSREG1_ROUT(SYSREG1_CLK_CTL, data);
	/* GbE clock enable */
	data = SYSREG1_RIN(SYSREG1_CLK_EN);
	data |= (SYSREG1_GBE_CKON | SYSREG1_GBEEXT_EN);
	SYSREG1_ROUT(SYSREG1_CLK_EN, data);

	return;
}
#endif /* CONFIG_XARINA */

/* phy register read */
static unsigned short smac_readphy(ushort reg)
{
	unsigned short result;

	smac_wait_phy();
	smac_writereg(SMAC_PHY_ADDR,
		(SMAC_PHY_LAYER_ADDR << SMAC_PA_S_PA) |
		SMAC_PHY_CLK |
		SMAC_PA_READ |
		SMAC_PA_KICK |
		((reg << SMAC_PA_S_GR) & SMAC_PA_M_GR));
	smac_wait_phy();
	result = smac_readreg(SMAC_PHY_DATA);

	return result;
}

/* phy register write */
static void smac_writephy(ushort reg, unsigned short data)
{
	smac_writereg(SMAC_PHY_DATA, data);

	smac_wait_phy();
	smac_writereg(SMAC_PHY_ADDR,
		(SMAC_PHY_LAYER_ADDR << SMAC_PA_S_PA) |
		SMAC_PHY_CLK  |
		SMAC_PA_WRITE |
		SMAC_PA_KICK  |
		((reg << SMAC_PA_S_GR) & SMAC_PA_M_GR));
	smac_wait_phy();

	return;
}


/********************************/
/* Tx and Rx setting            */
/********************************/
static void smac_rx_desc_init(struct eth_device* dev, unsigned long idx)
{
	struct smac_rx_res *rx = &((struct smac_resource *)dev->priv)->rx;
	invalidate_dcache_range((unsigned long)&rx->desc[idx], (unsigned long)&rx->desc[idx + 1]);
	/* flag clear	*/
	smac_write_desc(0, (void *)&(rx->desc[idx].srd_flags));	
	/* length#2	*/
	smac_write_desc(SMAC_FRAME_SIZE << SMAC_SRD_S_BUF_LEN2, (void *)&(rx->desc[idx].srd_length));
	/* don't use	*/
	smac_write_desc(0, (void *)&(rx->desc[idx].srd_addr1));	
	/* buffer#2	*/
	smac_write_desc((uint32_t)&rx->buff[idx][0], (void *)&(rx->desc[idx].srd_addr2));
	flush_dcache_range((unsigned long)&rx->desc[idx], (unsigned long)&rx->desc[idx + 1]);
	return;
}


static void smac_configure_rx(struct eth_device* dev)
{
	struct smac_rx_res *rx = &((struct smac_resource *)dev->priv)->rx;
	unsigned long tail;
	int i;

	/* rx resource init */
	rx->idx = 0;

	/* Setup the SMAC_RX_DB_BASE, SMAC_RX_DB_LENGTH_MINUS1 register */
	smac_writereg(SMAC_RX_DB_BASE, (unsigned long)&rx->desc[0]);
	smac_writereg(SMAC_RX_DB_LENGTH_MINUS1, SMAC_MAX_RXD - 1);


	/* all discriptor init */
	for(i = 0; i < SMAC_MAX_RXD; i++) {
		/* descriptor init */
		smac_rx_desc_init(dev, i);
	}

	/* last discriptor set in tail reg */
	tail = INDEX2TAIL_RX(SMAC_MAX_RXD - 1);
	smac_writereg(SMAC_RX_DB_TAIL, tail);		/* TAIL reg set (ALL descriptor) */

	return;
}


static void smac_configure_tx(struct eth_device* dev)
{
	struct smac_tx_res *tx = &((struct smac_resource *)dev->priv)->tx;

	/* tx resource init */
	tx->idx = 0;

	/* Setup the SMAC_TX_DB_BASE, SMAC_TX_DB_LENGTH_MINUS1 register */
	smac_writereg(SMAC_TX_DB_BASE, (unsigned long)&tx->desc[0]);
	smac_writereg(SMAC_TX_DB_LENGTH_MINUS1, SMAC_MAX_TXD - 1);

	return;
}


/********************************/
/* init/open setting            */
/********************************/
static int smac_init_busmode(void)
{
	
#ifdef SMAC_CPU_DESC_BE
	smac_writereg(SMAC_BUS_MODE,
		      (0 << SMAC_BM_S_DCT) | (SMAC_PBL << SMAC_BM_S_PBL) |
		      SMAC_BM_DBO_LITTLE | SMAC_BM_BLE_BIG);
#endif /* SMAC_CPU_DESC_BE */
#ifdef SMAC_CPU_DESC_LE
	smac_writereg(SMAC_BUS_MODE,
		      (0 << SMAC_BM_S_DCT) | (SMAC_PBL << SMAC_BM_S_PBL) |
		      SMAC_BM_DBO_LITTLE | SMAC_BM_BLE_LITTLE);
#endif /* SMAC_CPU_DESC_LE */

	
	return 0;
}


static void smac_reset(void)
{
	volatile uint32_t busmode = 0;
	volatile uint32_t mmc_control = 0;
	volatile uint32_t sw_status = 0;

	DEBUG_PRINTK2("smac_reset()\n");

	/* SW reset */
	busmode = smac_readreg(SMAC_BUS_MODE);
	smac_writereg(SMAC_BUS_MODE, busmode | SMAC_BM_SWR);

	/* wait reset is done */
	do {
		busmode = smac_readreg(SMAC_BUS_MODE);
		DEBUG_PRINTK2("busmode:0x%08x\n", busmode);

		sw_status = smac_readreg(SMAC_STATE_SWR);
		DEBUG_PRINTK2("sw_status:0x%08x\n", sw_status);
	} while (busmode & SMAC_BM_SWR);

	/* re-write busmode */
	/* because it is cleared by reset, the bus mode sets it again. */
	smac_init_busmode();

	/* stat counter cler */
	mmc_control = SMAC_MMC_CR;
	smac_writereg(SMAC_MMC_CONTROL, mmc_control);

	return;
}


static int smac_configure_mac(unsigned char *enetaddr)
{
	uint32_t addr_low, addr_high;

	/*
	 * MAC config
	 * set GMII, Full Duplex as default
	 * send wdt & jubber timer is disable.
	 * transmid & receive is enable.
	 */
	 
	smac_writereg(SMAC_MAC_CONFIG,
			SMAC_MC_WD	| SMAC_MC_JD	| SMAC_MC_FES_100	|
			SMAC_MC_DM_FULL	| SMAC_MC_TE	| SMAC_MC_RE		|
			SMAC_MC_DFS_ENA	| SMAC_MC_RCS	| SMAC_MC_PS_MII);

	/* MAC address */
	DEBUG_PRINTK("mac addr: %02x %02x %02x %02x %02x %02x\n",
		enetaddr[0], enetaddr[1],
		enetaddr[2], enetaddr[3],
		enetaddr[4], enetaddr[5]);

	addr_low  = (((uint32_t) enetaddr[0])       |
	             ((uint32_t) enetaddr[1] << 8)  |
	             ((uint32_t) enetaddr[2] << 16) |
	             ((uint32_t) enetaddr[3] << 24));
	addr_high = (((uint32_t) enetaddr[4])       |
	             ((uint32_t) enetaddr[5] << 8));

	smac_writereg(SMAC_MAC_ADDR_HIGH(0), addr_high | SMAC_MAC_AE);
	smac_writereg(SMAC_MAC_ADDR_LOW(0),  addr_low);

	DEBUG_PRINTK("mac addr: %08lx %08lx\n",
		     smac_readreg(SMAC_MAC_ADDR_HIGH(0)),
		     smac_readreg(SMAC_MAC_ADDR_LOW(0)));

	return 0;
}


/* multicast & promiscuous is disable in this function */
static void smac_set_multi(void)
{
	uint32_t framefilter;
	int i;

	/* multicast & promiscuous is disable*/
	framefilter = smac_readreg(SMAC_MAC_FRAME_FILTER);
	framefilter &= ~(SMAC_MFF_PR | SMAC_MFF_PM);

	smac_writereg(SMAC_MAC_FRAME_FILTER, framefilter);

	for (i = 1; i < SMAC_MULTICAST_ADDR_NUM; i++) {
		smac_writereg(SMAC_MAC_ADDR_HIGH(i), 0xffff);
		smac_writereg(SMAC_MAC_ADDR_LOW(i), ~0);
	}

	return;
}

static void smac_sw_init(struct eth_device* dev)
{
	/* software reset */
	smac_reset();

	/* interrupt disable */
	smac_writereg(SMAC_INTR_DISABLE, ~0);
	/* interrupt clear */
	smac_writereg(SMAC_INTR_CLEAR, ~0);

	/* SMAC_OPERAOIN register init */
	/* RxFIFO&TxFIFO is set in Store and Forward Mode set*/
	smac_writereg(SMAC_OPERATION, SMAC_OP_TSF | SMAC_OP_RSF);

	/* re-write busmode */
	smac_init_busmode();

	/* mac address set */
	smac_configure_mac(dev->enetaddr);

	return;
}

/********************************/
/* autonego                     */
/********************************/
static void smac_autonego_set(void)
{
	uint16_t control;

	/* set auto-neg advertisement register */
	control = smac_readphy(SMAC_PHY_AUTONEG_ADVERTISE);
	control |= (1 << 8) | (1 << 7) | (1 << 6) | (1 << 5);
	smac_writephy(SMAC_PHY_AUTONEG_ADVERTISE, control);

	/* set control register */
	control = (SMAC_PHY_S_100BASE_FULL | SMAC_PHY_S_AUTO_NEG_ENA
		 | SMAC_PHY_S_AUTO_NEG_RESTART);

	smac_writephy(SMAC_PHY_CONTROL, control);
	control = smac_readphy(SMAC_PHY_CONTROL);
	DEBUG_PRINTK2("control=0x%08x\n", control);


	return;
}


static int smac_get_speed_and_duplex(uint16_t *speed, uint16_t *duplex)
{
	uint32_t status;

	/* read phy status register */
	status = smac_readphy(SMAC_PHY_STATUS);
	DEBUG_PRINTK2("PHY::STATUS = 0x%08x\n", status);

	status = smac_readphy(SMAC_PHY_PHY_CONTROL1);
	DEBUG_PRINTK2("PHY::SPECIFIC_STATUS = 0x%08x\n", status);

	/* speed */
	switch (status & SMAC_PHY_M_SPEED) {
	case 0x2:
		*speed = SPEED_100;
		DEBUG_PRINTK("100 Mbps, ");
		break;
	case 0x1:
		*speed = SPEED_10;
		DEBUG_PRINTK("10 Mbps, ");
		break;
	default:
		*speed = SPEED_10;
		DEBUG_PRINTK("wrong speed?\n");
		return FALSE;
	}

	/* duplex */
	if (status & SMAC_PHY_M_DUPLEX) {
		*duplex = FULL_DUPLEX;
		DEBUG_PRINTK("Full Duplex\n");
	} else {
		*duplex = HALF_DUPLEX;
		DEBUG_PRINTK(" Half Duplex\n");
	}
	
	printf("MII mode\n");
	
	return TRUE;
}


static int smac_link_check(uint16_t *speed, uint16_t *duplex)
{
	uint32_t link;
	uint16_t link_speed, link_duplex;
	uint32_t config;

	link = smac_readphy(SMAC_PHY_PHY_CONTROL1) & SMAC_PHY_M_LINK;

	if (!link) {
		return FALSE;
	}

	if (!smac_get_speed_and_duplex(&link_speed, &link_duplex)) {
		return FALSE;
	}
	
	/* MAC config */
	config = smac_readreg(SMAC_MAC_CONFIG);
	/* Port Select */
	if (link_speed == SPEED_1000) {
		config &= ~SMAC_MC_PS_MII;
	} else {
		config |= SMAC_MC_PS_MII;
	}
	/* Fast Ethernet */
	if (link_speed == SPEED_10) {
		config &= ~SMAC_MC_FES_100;
	} else {
		config |= SMAC_MC_FES_100;
	}
	/* Duplex Mode */
	if (link_duplex == FULL_DUPLEX) {
		config |= SMAC_MC_DM_FULL;
	} else {
		config &= ~SMAC_MC_DM_FULL;
		config &= ~SMAC_MC_M_IFG;
		config |= SMAC_MC_IFG_64BIT;
	}
	smac_writereg(SMAC_MAC_CONFIG, config);

	/* PHY setting */
#if 0
	if (link_duplex == HALF_DUPLEX) {
		phy_ctrl = smac_readphy(SMAC_PHY_SPEC_CONTROL);
		phy_ctrl |= SMAC_PHY_S_ASSERT_CRS;
		smac_writephy(SMAC_PHY_SPEC_CONTROL, phy_ctrl);
	}
#endif

#ifdef CONFIG_XARINA
	/* SYSREG1 CLK_CTL setting */
	smac_write_clock_reg(SYSREG1_GBE_CKSEL_MII10M_100M);
#endif /* CONFIG_XARINA */

	/* value return */
	*speed = link_speed;
	*duplex = link_duplex;
	

	return TRUE;
}


/********************************/
/* up & down                    */
/********************************/
static int smac_up(struct eth_device* dev)
{
	uint16_t link_speed, link_duplex;	
	unsigned long start, now;
	int ret;

	DEBUG_PRINTK2("smac_up()\n");

	smac_set_multi();

	/* descriptori initialize */
	smac_configure_tx(dev);
	smac_configure_rx(dev);

	/* Transfer/Receive is enable */
	smac_writereg(SMAC_OPERATION,
		smac_readreg(SMAC_OPERATION) | SMAC_OP_ST | SMAC_OP_SR);

	/* auto negotiation exec */
	printf("start auto negotiation.\n");
	smac_autonego_set();

	/* link check */
	start = get_timer(0);
	while(1) {
		ret = smac_link_check(&link_speed, &link_duplex);
		if (ret) {
			udelay(500000);
			/* link OK */
			printf("auto negotiation complete,");
			switch(link_speed) {
			case SPEED_1000:
				printf("1000Mbps, ");
				break;
			case SPEED_100:
				printf("100Mbps, ");
				break;
			case SPEED_10:
				printf("10Mbps, ");
				break;
			default:
				break;
			}
			switch(link_duplex) {
			case FULL_DUPLEX:
				printf("full duplex\n");
				break;
			case HALF_DUPLEX:
				printf("half duplex\n");
				break;
			default:
				break;
			}

			break;
		}

		now = get_timer(start);
		if (now >= SMAC_LINK_TIMEOUT * get_tbclk()) {
			printf("auto negotiation failed.\n");
			break;
		}

	}

	return ret;
}


static void smac_down(void)
{
	DEBUG_PRINTK2("smac_down\n");

	/* software reset */
	smac_reset();

	/* Transfer/Receive is disable */
	smac_writereg(SMAC_OPERATION,
		smac_readreg(SMAC_OPERATION) & ~SMAC_OP_ST & ~SMAC_OP_SR);

	return;
}


/********************************/
/* I/F                          */
/********************************/
/* init */
/* return value: success(1), failed(0) */
static int smac_init(struct eth_device* dev, bd_t* bd)
{
	int ret;

	g_smac_regbase = dev->iobase;
	
	/* init */
	smac_sw_init(dev);
	/* open */
	printf("smac detected.\n");
	DEBUG_PRINTK2("smac version: 0x%08x\n", smac_readreg(SMAC_HW_VERSION));

	ret = smac_up(dev);

	return ret;
}

/* halt */
static void smac_halt(struct eth_device* dev)
{
	g_smac_regbase = dev->iobase;
	
	smac_down();

	return;
}


/* frame recieve */
/* return value: receive frame bytes num */
static int smac_recv(struct eth_device* dev)
{
	struct smac_rx_res *rx = &((struct smac_resource *)dev->priv)->rx;
	uint32_t srd_flags;
	uint32_t srd_length;
	unsigned long idx;
	unsigned long tail;
	int err = FALSE;

	g_smac_regbase = dev->iobase;

	idx = rx->idx;
	invalidate_dcache_range((unsigned long)&rx->desc[idx], (unsigned long)&rx->desc[idx + 1]);
	srd_flags = rx->desc[idx].srd_flags;

	/* frame receive check */
	if (!(srd_flags & SMAC_STDD_M_WB)) {
		/* none recieve frame */
		return 0;
	}

	/* descriptor write back data check */
	if (srd_flags & SMAC_SRD_M_ES) {
		/* receive error */
		printf("rx error.\n");
		DEBUG_PRINTK2("srd_flags:0x%08x\n", srd_flags);
		err = TRUE;
	}

	/* receive length check */
	srd_length = (rx->desc[idx].srd_flags & SMAC_SRD_M_FRAME_LEN) >> SMAC_SRD_S_FRAME_LEN;
	if (srd_length > SMAC_FRAME_SIZE) { 
		/* receive error */
		printf("rx frame size error.\n");
		DEBUG_PRINTK2("length:0x%08x\n", srd_length);
		err = TRUE;
	}


	DEBUG_PRINTK2("receive index :0x%08x\n", idx);
	DEBUG_PRINTK2("receive p_desc:0x%08x\n", &rx->desc[idx]);
	DEBUG_PRINTK2("receive buff  :0x%08x\n", rx->desc[idx].srd_addr2);
	DEBUG_PRINTK2("receive len   :0x%08x\n", srd_length);
	DEBUG_PRINTK2("head(addr)    :0x%08x\n", smac_readreg(SMAC_RX_DB_HEAD));
	DEBUG_PRINTK2("head(index)   :0x%08x\n", HEAD2INDEX_RX(smac_readreg(SMAC_RX_DB_HEAD)));
	DEBUG_PRINTK2("tail(addr)    :0x%08x\n", smac_readreg(SMAC_RX_DB_TAIL));
	DEBUG_PRINTK2("tail(index)   :0x%08x\n", TAIL2INDEX_RX(smac_readreg(SMAC_RX_DB_TAIL)));

	
	if (!err) {	/* receive OK */
		/* receive flame data to u-boot */
		invalidate_dcache_range(rx->desc[idx].srd_addr2, 
								rx->desc[idx].srd_addr2 + srd_length);
		NetReceive((volatile uchar *)(rx->desc[idx].srd_addr2), (int)srd_length);
	}

	/* receive discriptor init */
	smac_rx_desc_init(dev, idx);

	/* receive index set tail reg */
	tail = INDEX2TAIL_RX(idx);
	smac_writereg(SMAC_RX_DB_TAIL, tail);		/* TAIL reg set */

	/* next index move */
	rx->idx = INDEX_INC_RX(idx);

	DEBUG_PRINTK2("tail(addr)    :0x%08x\n", smac_readreg(SMAC_RX_DB_TAIL));
	DEBUG_PRINTK2("tail(index)   :0x%08x\n", TAIL2INDEX_RX(smac_readreg(SMAC_RX_DB_TAIL)));

	if (err) {	/* receive NG */
		srd_length = 0;
	}

	return srd_length;
}


/* frame send */
/* *packet     : packet data pointer  */
/* length      : packet data length   */
/* return value: send frame bytes num */
static int smac_send(struct eth_device* dev, volatile void* packet, int length)
{
	struct smac_tx_res *tx = &((struct smac_resource *)dev->priv)->tx;
	void *p_buff = &tx->buff[0][0];
	unsigned long state, flag;
	unsigned long start, now;

	g_smac_regbase = dev->iobase;

	/* length check */
	if (length > SMAC_FRAME_SIZE) {
		return 0;
	}

	/* frame data copy */
	memcpy(p_buff, (void *)packet, length);
	flush_dcache_range((unsigned long)p_buff, (unsigned long)(p_buff + length));

	/* descriptor set */
	invalidate_dcache_range((unsigned long)&tx->desc[0], (unsigned long)&tx->desc[1]);
	/* Last & First Discriptor	*/
	smac_write_desc(SMAC_STDD_M_LS | SMAC_STDD_M_FS, (void *)&(tx->desc[0].std_flags));	
	/* length#2	*/
	smac_write_desc(length << SMAC_STDD_S_BUF_LEN2, (void *)&(tx->desc[0].std_length));
	/* don't use	*/
	smac_write_desc(0, (void *)&(tx->desc[0].std_addr1));	
	/* buffer#2	*/
	smac_write_desc((uint32_t)p_buff, (void *)&(tx->desc[0].std_addr2));
	flush_dcache_range((unsigned long)&tx->desc[0], (unsigned long)&tx->desc[1]);
	/* 1frame send */
	smac_writereg(SMAC_TX_DB_TAIL, (unsigned long)&tx->desc[0]);	/* TAIL reg set (ALL descriptor) */

	DEBUG_PRINTK2("send desc addr :0x%08x\n", (unsigned long)&tx->desc[0]);
	DEBUG_PRINTK2("send buff addr :0x%08x\n", (unsigned long)p_buff);
	DEBUG_PRINTK2("send len    :0x%08x\n", length);

	/* send complete check */
	start = get_timer(0);
	while(1) {
		/* send discriptor none? */
		state = smac_readreg(SMAC_INTR_STAT) & SMAC_INTR_TU;	/* Transmit Buffer Unavailable */
		invalidate_dcache_range((unsigned long)&tx->desc[0], (unsigned long)&tx->desc[1]);
		flag  = tx->desc[0].std_flags & SMAC_STDD_M_WB;	/*  Tx Discriptor write back */
		if (state && flag) {
			/* send OK */
			DEBUG_PRINTK2("SMAC_INTR_STAT:0x%08x\n", smac_readreg(SMAC_INTR_STAT));
			break;
		}

		now = get_timer(start);
		if (now >= SMAC_TX_TIMEOUT * get_tbclk()) {
			printf("tx timeout error.\n");
			return 0;
		}

	}

	/* descriptor write back data check */
	if (tx->desc[0].std_flags & SMAC_STDD_M_ES) {
		/* send error */
		printf("tx error.\n");
		DEBUG_PRINTK("std_flags:0x%08x\n", tx->desc[0].std_flags);
		return 0;
	}

	DEBUG_PRINTK2("send ok\n");

	return length;
}

static int smac_write_hwaddr(struct eth_device* dev)
{
	/* mac address set */
	smac_configure_mac(dev->enetaddr);
	smac_sw_init(dev);
	return 0;
}

static unsigned long g_base[SMAC_MAX_DEVICE] =  {
	SMAC_BASE_00,
};

static struct eth_device g_device[SMAC_MAX_DEVICE];

/* driver initialize        */
/* return value: register device num */
int smac_mii_initialize(bd_t *bis)
{
	int i;
	struct eth_device *dev = NULL;

	for(i = 0; i < SMAC_MAX_DEVICE; i++) {
		dev = &g_device[i];
		memset(dev, 0, sizeof(struct eth_device));
		dev->init = smac_init;
		dev->halt = smac_halt;
		dev->recv = smac_recv;
		dev->send = smac_send;
		dev->write_hwaddr = smac_write_hwaddr;

		sprintf(dev->name, "smac#%d", i);
		dev->iobase = g_base[i];
		dev->priv   = (void *)malloc(sizeof(struct smac_resource));
		if (dev->priv == NULL) {
			printf("smac initialize error.\n");
			return 0;
		}

		DEBUG_PRINTK2("malloc addr :0x%08x\n", (unsigned long)dev->priv);

		/* regist g_device struct */
		eth_register(dev);

		/* initialize */
		g_smac_regbase = dev->iobase;

	}

 
	if (dev == NULL) {
		printf("smac initialize error.\n");
		return 0;
	}


	return i;
}

#endif /* defined(CONFIG_SMAC) && defined(CONFIG_NET_MULTI) */
